'''
See COPYRIGHT.md for copyright information.
'''
from arelle import ModelDocument, XmlUtil, XbrlConst
from arelle.ModelDtsObject import ModelConcept, ModelLocator, ModelResource, ModelType
from arelle.ModelValue import qname
from arelle.ModelObject import ModelObject, ModelComment
from arelle.ValidateXbrlDTS import arcFromConceptQname, arcToConceptQname
from arelle import XmlValidate
from lxml import etree
import regex as re

xsd1_1datatypes = {qname(XbrlConst.xsd,'anyAtomicType'), qname(XbrlConst.xsd,'yearMonthDuration'), qname(XbrlConst.xsd,'dayTimeDuration'), qname(XbrlConst.xsd,'dateTimeStamp'), qname(XbrlConst.xsd,'precisionDecimal')}

def checkDTSdocument(val, modelDocument, isFilingDocument):
    if not isFilingDocument:
        return  # not a filing's extension document

    val.valUsedPrefixes = set()
    val.referencedNamespaces = set()
    val.substititutionGroupQname = {}

    modelXbrl = val.modelXbrl
    if modelDocument.type in (ModelDocument.Type.SCHEMA, ModelDocument.Type.LINKBASE):
        isSchema = modelDocument.type == ModelDocument.Type.SCHEMA
        docinfo = modelDocument.xmlDocument.docinfo
        if docinfo and docinfo.xml_version != "1.0":
            modelXbrl.error("SBR.NL.2.2.0.02" if isSchema else "SBR.NL.2.3.0.02",
                    _('%(docType)s xml version must be "1.0" but is "%(xmlVersion)s"'),
                    modelObject=modelDocument, docType=modelDocument.gettype().title(),
                    xmlVersion=docinfo.xml_version)
        if modelDocument.documentEncoding.lower() != "utf-8":
            modelXbrl.error("SBR.NL.2.2.0.03" if isSchema else "SBR.NL.2.3.0.03",
                    _('%(docType)s encoding must be "utf-8" but is "%(xmlEncoding)s"'),
                    modelObject=modelDocument, docType=modelDocument.gettype().title(),
                    xmlEncoding=modelDocument.documentEncoding)
        lookingForPrecedingComment = True
        for commentNode in modelDocument.xmlRootElement.itersiblings(preceding=True):
            if isinstance(commentNode, etree._Comment):
                if lookingForPrecedingComment:
                    lookingForPrecedingComment = False
                else:
                    modelXbrl.error("SBR.NL.2.2.0.05" if isSchema else "SBR.NL.2.3.0.05",
                            _('%(docType)s must have only one comment node before schema element'),
                            modelObject=modelDocument, docType=modelDocument.gettype().title())
        if lookingForPrecedingComment:
            modelXbrl.error("SBR.NL.2.2.0.04" if isSchema else "SBR.NL.2.3.0.04",
                _('%(docType)s must have comment node only on line 2'),
                modelObject=modelDocument, docType=modelDocument.gettype().title())

        if isSchema:
            for elt in modelDocument.xmlRootElement.iter(tag="{http://www.w3.org/2001/XMLSchema}*"):
                parent = elt.getparent()
                parentIsSchema = parent is not None and parent.tag == "{http://www.w3.org/2001/XMLSchema}schema"
                localName = elt.localName
                if localName == "schema":
                    if elt.get("targetNamespace") is None:
                        val.modelXbrl.error("SBR.NL.2.2.0.08",
                            _('Schema element must have a targetNamespace attribute'),
                            modelObject=elt)
                    if (elt.get("attributeFormDefault") != "unqualified" or
                        elt.get("elementFormDefault") != "qualified"):
                        val.modelXbrl.error("SBR.NL.2.2.0.09",
                                _('Schema element attributeFormDefault must be "unqualified" and elementFormDefault must be "qualified"'),
                                modelObject=elt)
                    for attrName in ("blockDefault", "finalDefault", "version"):
                        if elt.get(attrName) is not None:
                            val.modelXbrl.error("SBR.NL.2.2.0.10",
                                _('Schema element must not have a %(attribute)s attribute'),
                                modelObject=elt, attribute=attrName)
                else:
                    if localName in ("assert", "openContent", "fallback"):
                        val.modelXbrl.error("SBR.NL.2.2.0.01",
                            _('Schema contains XSD 1.1 content "%(element)s"'),
                            modelObject=elt, element=elt.qname)

                    if localName == "element":
                        for attr, presence, errCode in (("block", False, "2.2.2.09"),
                                                        ("final", False, "2.2.2.10"),
                                                        ("fixed", False, "2.2.2.11"),
                                                        ("form", False, "2.2.2.12"),):
                            if (elt.get(attr) is not None) != presence:
                                val.modelXbrl.error("SBR.NL.{0}".format(errCode),
                                    _('Schema element %(concept)s %(requirement)s contain attribute %(attribute)s'),
                                    modelObject=elt, concept=elt.get("name"),
                                    requirement=(_("MUST NOT"),_("MUST"))[presence], attribute=attr,
                                    messageCodes=("SBR.NL.2.2.2.09", "SBR.NL.2.2.2.10", "SBR.NL.2.2.2.11", "SBR.NL.2.2.2.12"))
                        eltName = elt.get("name")
                        if eltName is not None: # skip for concepts which are refs
                            type = qname(elt, elt.get("type"))
                            eltQname = elt.qname
                            if type in xsd1_1datatypes:
                                val.modelXbrl.error("SBR.NL.2.2.0.01",
                                    _('Schema element %(concept)s contains XSD 1.1 datatype "%(xsdType)s"'),
                                    modelObject=elt, concept=elt.get("name"), xsdType=type)
                            if not parentIsSchema: # root element
                                if elt.get("name") is not None and (elt.isItem or elt.isTuple):
                                    val.modelXbrl.error("SBR.NL.2.2.2.01",
                                        _('Schema concept definition is not at the root level: %(concept)s'),
                                        modelObject=elt, concept=elt.get("name"))
                            elif eltQname not in val.typedDomainQnames:
                                for attr, presence, errCode in (("abstract", True, "2.2.2.08"),
                                                                ("id", True, "2.2.2.13"),
                                                                ("nillable", True, "2.2.2.15"),
                                                                ("substitutionGroup", True, "2.2.2.18"),):
                                    if (elt.get(attr) is not None) != presence:
                                        val.modelXbrl.error("SBR.NL.{0}".format(errCode),
                                            _('Schema root element %(concept)s %(requirement)s contain attribute %(attribute)s'),
                                            modelObject=elt, concept=elt.get("name"),
                                            requirement=(_("MUST NOT"),_("MUST"))[presence], attribute=attr,
                                            messageCodes=("SBR.NL.2.2.2.08", "SBR.NL.2.2.2.13", "SBR.NL.2.2.2.15", "SBR.NL.2.2.2.18"))
                            # semantic checks
                            if elt.isTuple:
                                val.hasTuple = True
                            elif elt.isLinkPart:
                                val.hasLinkPart = True
                            elif elt.isItem:
                                if elt.isDimensionItem:
                                    val.hasDimension = True
                                #elif elt.substitutesFor()
                                if elt.isAbstract:
                                    val.hasAbstractItem = True
                                else:
                                    val.hasNonAbstraceElement = True
                            if elt.isAbstract and elt.isItem:
                                val.hasAbstractItem = True
                            if elt.typeQname is not None:
                                val.referencedNamespaces.add(elt.typeQname.namespaceURI)
                            if elt.substitutionGroupQname is not None:
                                val.referencedNamespaces.add(elt.substitutionGroupQname.namespaceURI)
                            if elt.isTypedDimension and elt.typedDomainElement is not None:
                                val.referencedNamespaces.add(elt.typedDomainElement.namespaceURI)
                        else:
                            referencedElt = elt.dereference()
                            if referencedElt is not None:
                                val.referencedNamespaces.add(referencedElt.modelDocument.targetNamespace)
                        if not parentIsSchema:
                            eltDecl = elt.dereference()
                            if (elt.get("minOccurs") is None or elt.get("maxOccurs") is None):
                                val.modelXbrl.error("SBR.NL.2.2.2.14",
                                    _('Schema %(element)s must have minOccurs and maxOccurs'),
                                    modelObject=elt, element=eltDecl.qname)
                            elif elt.get("maxOccurs") != "1" and eltDecl.isItem:
                                val.modelXbrl.error("SBR.NL.2.2.2.30",
                                    _("Tuple concept %(concept)s must have maxOccurs='1'"),
                                    modelObject=elt, concept=eltDecl.qname)
                            if eltDecl.isItem and eltDecl.isAbstract:
                                val.modelXbrl.error("SBR.NL.2.2.2.31",
                                    _("Abstract concept %(concept)s must not be a child of a tuple"),
                                    modelObject=elt, concept=eltDecl.qname)
                    elif localName in ("sequence","choice"):
                        for attrName in ("minOccurs", "maxOccurs"):
                            attrValue = elt.get(attrName)
                            if  attrValue is None:
                                val.modelXbrl.error("SBR.NL.2.2.2.14",
                                    _('Schema %(element)s must have %(attrName)s'),
                                    modelObject=elt, element=elt.elementQname, attrName=attrName)
                            elif attrValue != "1":
                                val.modelXbrl.error("SBR.NL.2.2.2.33",
                                    _('Schema %(element)s must have %(attrName)s = "1"'),
                                    modelObject=elt, element=elt.elementQname, attrName=attrName)
                    elif localName in {"complexType","simpleType"}:
                        qnameDerivedFrom = elt.qnameDerivedFrom
                        if qnameDerivedFrom is not None:
                            if isinstance(qnameDerivedFrom, list): # union
                                for qn in qnameDerivedFrom:
                                    val.referencedNamespaces.add(qn.namespaceURI)
                            else: # not union type
                                val.referencedNamespaces.add(qnameDerivedFrom.namespaceURI)
                    elif localName == "attribute":
                        if elt.typeQname is not None:
                            val.referencedNamespaces.add(elt.typeQname.namespaceURI)
                    elif localName == "appinfo":
                        if (parent.localName != "annotation" or parent.namespaceURI != XbrlConst.xsd or
                            parent.getparent().localName != "schema" or parent.getparent().namespaceURI != XbrlConst.xsd or
                            XmlUtil.previousSiblingElement(parent) != None):
                            val.modelXbrl.error("SBR.NL.2.2.0.12",
                                _('Annotation/appinfo record must be be behind schema and before import'), modelObject=elt)
                        nextSiblingElement = XmlUtil.nextSiblingElement(parent)
                        if nextSiblingElement is not None and nextSiblingElement.localName != "import":
                            val.modelXbrl.error("SBR.NL.2.2.0.14",
                                _('Annotation/appinfo record must be followed only by import'),
                                modelObject=elt)
                    elif localName == "annotation":
                        val.annotationsCount += 1
                        if not XmlUtil.hasChild(elt,XbrlConst.xsd,"appinfo"):
                            val.modelXbrl.error("SBR.NL.2.2.0.12",
                                _('Schema file annotation missing appinfo element must be be behind schema and before import'),
                                modelObject=elt)
                    elif localName in {"all", "documentation", "any", "anyAttribute", "attributeGroup",
                                       # comment out per R.H. 2011-11-16 "complexContent", "complexType", "extension",
                                       "field", "group", "key", "keyref",
                                       "list", "notation", "redefine", "selector", "unique"}:
                        val.modelXbrl.error("SBR.NL.2.2.11.{0:02}".format({"all":1, "documentation":2, "any":3, "anyAttribute":4, "attributeGroup":7,
                                                                  "complexContent":10, "complexType":11, "extension":12, "field":13, "group":14, "key":15, "keyref":16,
                                                                  "list":17, "notation":18, "redefine":20, "selector":22, "unique":23}[localName]),
                            _('Schema file element must not be used "%(element)s"'),
                            modelObject=elt, element=elt.qname,
                            messageCodes=("SBR.NL.2.2.11.01", "SBR.NL.2.2.11.02", "SBR.NL.2.2.11.03", "SBR.NL.2.2.11.04", "SBR.NL.2.2.11.07", "SBR.NL.2.2.11.10", "SBR.NL.2.2.11.11", "SBR.NL.2.2.11.12",
                                          "SBR.NL.2.2.11.13", "SBR.NL.2.2.11.14", "SBR.NL.2.2.11.15", "SBR.NL.2.2.11.16", "SBR.NL.2.2.11.17", "SBR.NL.2.2.11.18", "SBR.NL.2.2.11.20", "SBR.NL.2.2.11.22", "SBR.NL.2.2.11.23"))
                if not elt.prefix:
                        val.modelXbrl.error("SBR.NL.2.2.0.06",
                                'Schema element is not prefixed: "%(element)s"',
                                modelObject=elt, element=elt.qname)
            for elt in modelDocument.xmlRootElement.iter(tag="{http://www.xbrl.org/2003/linkbase}*"):
                if elt.localName in ("roleType","arcroleType"):
                    uriAttr = {"roleType":"roleURI",
                               "arcroleType":"arcroleURI"}[elt.localName]
                    XmlValidate.validate(val.modelXbrl, elt) # validate [arc]roleType
                    roleURI = elt.get(uriAttr)
                    if elt.localName == "arcroleType":
                        val.modelXbrl.error("SBR.NL.2.2.4.01",
                                _('ArcroleType is not allowed %(roleURI)s'),
                                modelObject=elt, roleURI=roleURI)
                    else: # roleType
                        roleTypeModelObject = modelDocument.idObjects.get(elt.get("id"))
                        if roleTypeModelObject is not None and not roleTypeModelObject.genLabel(lang="nl"):
                            val.modelXbrl.error("SBR.NL.2.3.8.05",
                                _('RoleType %(roleURI)s must have a label in lang "nl"'),
                                modelObject=elt, roleURI=roleURI)
                    # check for used on duplications
                    usedOns = set()
                    for usedOn in elt.iterdescendants(tag="{http://www.xbrl.org/2003/linkbase}usedOn"):
                        if isinstance(usedOn,ModelObject):
                            qName = qname(usedOn, XmlUtil.text(usedOn))
                            val.valUsedPrefixes.add(qName.prefix)
                            if qName == XbrlConst.qnLinkCalculationLink:
                                val.modelXbrl.error("SBR.NL.2.2.3.01",
                                    _("%(element)s usedOn must not be link:calculationLink"),
                                    modelObject=elt, element=elt.qname, value=qName)
                            if elt.localName == "roleType" and qName in XbrlConst.standardExtLinkQnames:
                                if not any((key[1] == roleURI  and key[2] == qName)
                                           for key in val.modelXbrl.baseSets.keys()):
                                    val.modelXbrl.error("SBR.NL.2.2.3.02",
                                        _("%(element)s usedOn %(usedOn)s not addressed for role %(role)s"),
                                        modelObject=elt, element=elt.qname, usedOn=qName, role=roleURI)
            if val.annotationsCount > 1:
                modelXbrl.error("SBR.NL.2.2.0.22",
                    _('Schema has %(annotationsCount)s xs:annotation elements, only 1 allowed'),
                    modelObject=modelDocument, annotationsCount=val.annotationsCount)
    if modelDocument.type == ModelDocument.Type.LINKBASE:
        for elt in modelDocument.xmlRootElement.iter():
            if isinstance(elt, ModelObject) and not elt.prefix:
                val.modelXbrl.error("SBR.NL.2.2.0.06",
                    _('Linkbase element is not prefixed: "%(element)s"'),
                    modelObject=elt, element=elt.qname)
        if not val.containsRelationship:
            modelXbrl.error("SBR.NL.2.3.0.12",
                "Linkbase has no relationships",
                modelObject=modelDocument)
        # check file name suffixes
        extLinkElt = XmlUtil.descendant(modelDocument.xmlRootElement, XbrlConst.link, "*", "{http://www.w3.org/1999/xlink}type", "extended")
        if extLinkElt is not None:
            expectedSuffix = None
            if extLinkElt.localName == "labelLink":
                anyLabel = XmlUtil.descendant(modelDocument.xmlRootElement, XbrlConst.link, "label", "{http://www.w3.org/XML/1998/namespace}lang", "*")
                if anyLabel is not None:
                    xmlLang = anyLabel.get("{http://www.w3.org/XML/1998/namespace}lang")
                    expectedSuffix = "-lab-{0}.xml".format(xmlLang)
            else:
                expectedSuffix = {"referenceLink": "-ref.xml",
                                  "presentationLink": "-pre.xml",
                                  "definitionLink": "-def.xml"}.get(extLinkElt.localName, None)
            if expectedSuffix:
                if not modelDocument.uri.endswith(expectedSuffix):
                    modelXbrl.error("SBR.NL.3.2.1.09",
                        "Linkbase filename expected to end with %(expectedSuffix)s: %(filename)s",
                        modelObject=modelDocument, expectedSuffix=expectedSuffix, filename=modelDocument.uri)
            elif extLinkElt.qname == XbrlConst.qnGenLink:
                anyLabel = XmlUtil.descendant(modelDocument.xmlRootElement, XbrlConst.link, "label", "{http://www.w3.org/XML/1998/namespace}lang", "*")
                if anyLabel is not None:
                    xmlLang = anyLabel.get("{http://www.w3.org/XML/1998/namespace}lang")
                    expectedSuffix = "-generic-lab-{0}.xml".format(xmlLang)
                elif XmlUtil.descendant(modelDocument.xmlRootElement, XbrlConst.link, "reference") is not None:
                    expectedSuffix = "-generic-ref.xml"
                if expectedSuffix and not modelDocument.uri.endswith(expectedSuffix):
                    modelXbrl.error("SBR.NL.3.2.1.10",
                        "Generic linkbase filename expected to end with %(expectedSuffix)s: %(filename)s",
                        modelObject=modelDocument, expectedSuffix=expectedSuffix, filename=modelDocument.uri)
        # label checks
        for qnLabel in (XbrlConst.qnLinkLabel, XbrlConst.qnGenLabel):
            for modelLabel in modelDocument.xmlRootElement.iterdescendants(tag=qnLabel.clarkNotation):
                if isinstance(modelLabel, ModelResource):
                    if not modelLabel.text or not modelLabel.text[:1].isupper():
                        modelXbrl.error("SBR.NL.3.2.7.05",
                            _("Labels MUST have a capital first letter, label %(label)s: %(text)s"),
                            modelObject=modelLabel, label=modelLabel.xlinkLabel, text=modelLabel.text[:64])
                    if modelLabel.role in (XbrlConst.standardLabel, XbrlConst.genStandardLabel):
                        if len(modelLabel.text) > 255:
                            modelXbrl.error("SBR.NL.3.2.7.06",
                                _("Labels with the 'standard' role MUST NOT exceed 255 characters, label %(label)s: %(text)s"),
                                modelObject=modelLabel, label=modelLabel.xlinkLabel, text=modelLabel.text[:64])
                if modelLabel.role in (XbrlConst.standardLabel, XbrlConst.genStandardLabel):
                    if len(modelLabel.text) > 255:
                        modelXbrl.error("SBR.NL.3.2.7.06",
                            _("Labels with the 'standard' role MUST NOT exceed 255 characters, label %(label)s: %(text)s"),
                            modelObject=modelLabel, label=modelLabel.xlinkLabel, text=modelLabel.text[:64])
        for modelResource in modelDocument.xmlRootElement.iter():
            # locator checks
            if isinstance(modelResource, ModelLocator):
                hrefModelObject = modelResource.dereference()
                if isinstance(hrefModelObject, ModelObject):
                    expectedLocLabel = hrefModelObject.id + "_loc"
                    if modelResource.xlinkLabel != expectedLocLabel:
                        modelXbrl.error("SBR.NL.3.2.11.01",
                            _("Locator @xlink:label names MUST be concatenated from: @id from the XML node, underscore, 'loc', expected %(expectedLocLabel)s, found %(foundLocLabel)s"),
                            modelObject=modelResource, expectedLocLabel=expectedLocLabel, foundLocLabel=modelResource.xlinkLabel)
            # xlinkLabel checks
            if isinstance(modelResource, ModelResource):
                if re.match(r"[^a-zA-Z0-9_-]", modelResource.xlinkLabel):
                    modelXbrl.error("SBR.NL.3.2.11.03",
                        _("@xlink:label names MUST use a-zA-Z0-9_- characters only: %(xlinkLabel)s"),
                        modelObject=modelResource, xlinkLabel=modelResource.xlinkLabel)
    elif modelDocument.targetNamespace: # SCHEMA with targetNamespace
        # check for unused imports
        for referencedDocument in modelDocument.referencesDocument.keys():
            if (referencedDocument.type == ModelDocument.Type.SCHEMA and
                referencedDocument.targetNamespace not in {XbrlConst.xbrli, XbrlConst.link} and
                referencedDocument.targetNamespace not in val.referencedNamespaces):
                modelXbrl.error("SBR.NL.2.2.0.15",
                    _("A schema import schemas of which no content is being addressed: %(importedFile)s"),
                    modelObject=modelDocument, importedFile=referencedDocument.basename)
        if modelDocument.targetNamespace != modelDocument.targetNamespace.lower():
            modelXbrl.error("SBR.NL.3.2.3.02",
                _("Namespace URI's MUST be lower case: %(namespaceURI)s"),
                modelObject=modelDocument, namespaceURI=modelDocument.targetNamespace)
        if len(modelDocument.targetNamespace) > 255:
            modelXbrl.error("SBR.NL.3.2.3.03",
                _("Namespace URI's MUST NOT be longer than 255 characters: %(namespaceURI)s"),
                modelObject=modelDocument, namespaceURI=modelDocument.targetNamespace)
        if re.match(r"[^a-z0-9_/-]", modelDocument.targetNamespace):
            modelXbrl.error("SBR.NL.3.2.3.04",
                _("Namespace URI's MUST use only signs from a-z0-9_-/: %(namespaceURI)s"),
                modelObject=modelDocument, namespaceURI=modelDocument.targetNamespace)
        if not modelDocument.targetNamespace.startswith('http://www.nltaxonomie.nl'):
            modelXbrl.error("SBR.NL.3.2.3.05",
                _("Namespace URI's MUST start with 'http://www.nltaxonomie.nl': %(namespaceURI)s"),
                modelObject=modelDocument, namespaceURI=modelDocument.targetNamespace)
        namespacePrefix = XmlUtil.xmlnsprefix(modelDocument.xmlRootElement, modelDocument.targetNamespace)
        if not namespacePrefix:
            modelXbrl.error("SBR.NL.3.2.4.01",
                _("TargetNamespaces MUST have a prefix: %(namespaceURI)s"),
                modelObject=modelDocument, namespaceURI=modelDocument.targetNamespace)
        elif namespacePrefix in val.prefixNamespace:
            modelXbrl.error("SBR.NL.3.2.4.02",
                _("Namespace prefix MUST be unique within the NT but prefix '%(prefix)s' is used by both %(namespaceURI)s and %(namespaceURI2)s."),
                modelObject=modelDocument, prefix=namespacePrefix,
                namespaceURI=modelDocument.targetNamespace, namespaceURI2=val.prefixNamespace[namespacePrefix])
        else:
            val.prefixNamespace[namespacePrefix] = modelDocument.targetNamespace
            val.namespacePrefix[modelDocument.targetNamespace] = namespacePrefix
        if namespacePrefix in {"xsi", "xsd", "xs", "xbrli", "link", "xlink", "xbrldt", "xbrldi", "gen", "xl"}:
            modelXbrl.error("SBR.NL.3.2.4.03",
                _("Namespace prefix '%(prefix)s' reserved by organizations for international specifications is used %(namespaceURI)s."),
                modelObject=modelDocument, prefix=namespacePrefix, namespaceURI=modelDocument.targetNamespace)
        if len(namespacePrefix) > 20:
            modelXbrl.warning("SBR.NL.3.2.4.06",
                _("Namespace prefix '%(prefix)s' SHOULD not exceed 20 characters %(namespaceURI)s."),
                modelObject=modelDocument, prefix=namespacePrefix, namespaceURI=modelDocument.targetNamespace)
        # check every non-targetnamespace prefix against its schema
    requiredLinkrole = None # only set for extension taxonomies

    for elt in modelDocument.xmlRootElement.iter():
        if isinstance(elt, ModelObject):
            xlinkType = elt.get("{http://www.w3.org/1999/xlink}type")
            xlinkRole = elt.get("{http://www.w3.org/1999/xlink}role")
            if elt.tag in ("{http://www.xbrl.org/2003/linkbase}arcroleRef", "{http://www.xbrl.org/2003/linkbase}roleRef"):
                if elt.tag =="{http://www.xbrl.org/2003/linkbase}arcroleRef": # corrected merge of pre-plugin code per LOGIUS
                    for attrName, errCode in (("{http://www.w3.org/1999/xlink}arcrole","SBR.NL.2.3.2.05"),("{http://www.w3.org/1999/xlink}role","SBR.NL.2.3.2.06")):
                        if elt.get(attrName):
                            val.modelXbrl.error(errCode,
                                _("Arcrole %(refURI)s arcroleRef %(xlinkHref)s must not have an %(attribute)s attribute"),
                                modelObject=elt, refURI=elt.get("arcroleURI"),
                                xlinkHref=elt.get("{http://www.w3.org/1999/xlink}href"), attribute=attrName,
                                messageCodes=("SBR.NL.2.3.2.05", "SBR.NL.2.3.2.06"))
                elif elt.tag == "{http://www.xbrl.org/2003/linkbase}roleRef": # corrected merge of pre-plugin code per LOGIUS
                    for attrName, errCode in (("{http://www.w3.org/1999/xlink}arcrole","SBR.NL.2.3.10.09"),("{http://www.w3.org/1999/xlink}role","SBR.NL.2.3.10.10")):
                        if elt.get(attrName):
                            val.modelXbrl.error(errCode,
                                _("Role %(refURI)s roleRef %(xlinkHref)s must not have an %(attribute)s attribute"),
                                modelObject=elt, refURI=elt.get("roleURI"),
                                xlinkHref=elt.get("{http://www.w3.org/1999/xlink}href"), attribute=attrName,
                                messageCodes=("SBR.NL.2.3.10.09", "SBR.NL.2.3.10.10"))
                if not xlinkType:
                    val.modelXbrl.error("SBR.NL.2.3.0.01",
                        _("Xlink 1.1 simple type is not allowed (xlink:type is missing)"),
                        modelObject=elt)
            if xlinkType == "resource":
                if elt.localName in ("label", "reference"):
                    if not XbrlConst.isStandardRole(xlinkRole):
                        val.modelXbrl.error("SBR.NL.2.3.10.13",
                            _("Extended link %(element)s must have a standard xlink:role attribute (%(xlinkRole)s)"),
                            modelObject=elt, element=elt.elementQname, xlinkRole=xlinkRole)
                if elt.localName == "reference": # look for custom reference parts
                    for linkPart in elt.iterchildren():
                        if linkPart.namespaceURI not in val.disclosureSystem.baseTaxonomyNamespaces:
                            val.modelXbrl.error("SBR.NL.2.2.5.01",
                                _("Link part %(element)s is not authorized"),
                                modelObject=linkPart, element=linkPart.elementQname)
                # corrected merge of pre-plugin code per LOGIUS
                if not XbrlConst.isStandardRole(xlinkRole):
                    modelsRole = val.modelXbrl.roleTypes.get(xlinkRole)
                    if (modelsRole is None or len(modelsRole) == 0 or
                        modelsRole[0].modelDocument.targetNamespace not in val.disclosureSystem.standardTaxonomiesDict):
                        val.modelXbrl.error("SBR.NL.2.3.10.14",
                            _("Resource %(xlinkLabel)s role %(role)s is not a standard taxonomy role"),
                            modelObject=elt, xlinkLabel=elt.get("{http://www.w3.org/1999/xlink}label"), role=xlinkRole, element=elt.qname,
                            roleDefinition=val.modelXbrl.roleTypeDefinition(xlinkRole))
                if elt.localName == "reference":
                    for child in elt.iterdescendants():
                        if isinstance(child,ModelObject) and child.namespaceURI.startswith("http://www.xbrl.org") and child.namespaceURI != "http://www.xbrl.org/2006/ref":
                            val.modelXbrl.error("SBR.NL.2.3.3.01",
                                _("Reference %(xlinkLabel)s has unauthorized part element %(element)s"),
                                modelObject=elt, xlinkLabel=elt.get("{http://www.w3.org/1999/xlink}label"),
                                element=qname(child))
                    id = elt.get("id")
                    if not id:
                        val.modelXbrl.error("SBR.NL.2.3.3.02",
                            _("Reference %(xlinkLabel)s is missing an id attribute"),
                            modelObject=elt, xlinkLabel=elt.get("{http://www.w3.org/1999/xlink}label"))
                    elif id in val.DTSreferenceResourceIDs:
                        val.modelXbrl.error("SBR.NL.2.3.3.03",
                            _("Reference %(xlinkLabel)s has duplicated id %(id)s also in linkbase %(otherLinkbase)s"),
                            modelObject=elt, xlinkLabel=elt.get("{http://www.w3.org/1999/xlink}label"),
                            id=id, otherLinkbase=val.DTSreferenceResourceIDs[id])
                    else:
                        val.DTSreferenceResourceIDs[id] = modelDocument.basename
                if elt.qname not in {
                    XbrlConst.qnLinkLabelLink: (XbrlConst.qnLinkLabel,),
                    XbrlConst.qnLinkReferenceLink: (XbrlConst.qnLinkReference,),
                    XbrlConst.qnLinkPresentationLink: tuple(),
                    XbrlConst.qnLinkCalculationLink: tuple(),
                    XbrlConst.qnLinkDefinitionLink: tuple(),
                    XbrlConst.qnLinkFootnoteLink: (XbrlConst.qnLinkFootnote,),
                    # XbrlConst.qnGenLink: (XbrlConst.qnGenLabel, XbrlConst.qnGenReference, val.qnSbrLinkroleorder),
                     }.get(val.extendedElementName,(elt.qname,)):  # allow non-2.1 to be ok regardless per RH 2013-03-13
                    val.modelXbrl.error("SBR.NL.2.3.0.11",
                        _("Resource element %(element)s may not be contained in a linkbase with %(element2)s"),
                        modelObject=elt, element=elt.qname, element2=val.extendedElementName)
            elif xlinkType == "extended":
                if xlinkRole is None: # no @role on extended link
                    val.modelXbrl.error("SBR.NL.2.3.10.13",
                        _("Extended link %(element)s must have an xlink:role attribute"),
                        modelObject=elt, element=elt.elementQname)
                if not val.extendedElementName:
                    val.extendedElementName = elt.qname
                elif val.extendedElementName != elt.qname:
                    val.modelXbrl.error("SBR.NL.2.3.0.11",
                        _("Extended element %(element)s must be the same as %(element2)s"),
                        modelObject=elt, element=elt.qname, element2=val.extendedElementName)
            elif xlinkType == "locator":
                if elt.qname != XbrlConst.qnLinkLoc:
                    val.modelXbrl.error("SBR.NL.2.3.0.11",
                        _("Loc element %(element)s may not be contained in a linkbase with %(element2)s"),
                        modelObject=elt, element=elt.qname, element2=val.extendedElementName)
            elif xlinkType == "arc":
                if elt.namespaceURI == XbrlConst.link:
                    if elt.localName == "presentationArc" and not elt.get("order"):
                        val.modelXbrl.error("SBR.NL.2.3.4.04",
                            _("PresentationArc from %(xlinkFrom)s to %(xlinkTo)s must have an order"),
                            modelObject=elt,
                            xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                            xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                            conceptFrom=arcFromConceptQname(elt),
                            conceptTo=arcToConceptQname(elt))
                if elt.namespaceURI == XbrlConst.link:
                    arcrole = elt.get("{http://www.w3.org/1999/xlink}arcrole")
                    if elt.localName == "definitionArc":
                        if arcrole in (XbrlConst.essenceAlias, XbrlConst.similarTuples, XbrlConst.requiresElement):
                            val.modelXbrl.error({XbrlConst.essenceAlias: "SBR.NL.2.3.2.02",
                                              XbrlConst.similarTuples: "SBR.NL.2.3.2.03",
                                              XbrlConst.requiresElement: "SBR.NL.2.3.2.04"}[arcrole],
                                _("DefinitionArc from %(xlinkFrom)s to %(xlinkTo)s has unauthorized arcrole %(arcrole)s"),
                                modelObject=elt,
                                xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                                xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                                arcrole=arcrole,
                                messageCodes=("SBR.NL.2.3.2.02", "SBR.NL.2.3.2.03", "SBR.NL.2.3.2.04"))
                    elif elt.localName == "referenceArc":
                        if elt.get("order"):
                            val.modelXbrl.error("SBR.NL.2.3.3.05",
                                _("ReferenceArc from %(xlinkFrom)s to %(xlinkTo)s has an order"),
                                modelObject=elt,
                                xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                                xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"))
                    if elt.get("use") == "prohibited" and elt.getparent().tag in (
                            "{http://www.xbrl.org/2003/linkbase}presentationLink",
                            "{http://www.xbrl.org/2003/linkbase}labelLink",
                            "{http://xbrl.org/2008/generic}link",
                            "{http://www.xbrl.org/2003/linkbase}referenceLink"):
                        val.modelXbrl.error("SBR.NL.2.3.0.10",
                            _("%(arc)s must not contain use='prohibited'"),
                            modelObject=elt, arc=elt.getparent().qname)
                if elt.qname not in {
                    XbrlConst.qnLinkLabelLink: (XbrlConst.qnLinkLabelArc,),
                    XbrlConst.qnLinkReferenceLink: (XbrlConst.qnLinkReferenceArc,),
                    XbrlConst.qnLinkPresentationLink: (XbrlConst.qnLinkPresentationArc,),
                    XbrlConst.qnLinkCalculationLink: (XbrlConst.qnLinkCalculationArc,),
                    XbrlConst.qnLinkDefinitionLink: (XbrlConst.qnLinkDefinitionArc,),
                    XbrlConst.qnLinkFootnoteLink: (XbrlConst.qnLinkFootnoteArc,),
                    # XbrlConst.qnGenLink: (XbrlConst.qnGenArc,),
                     }.get(val.extendedElementName, (elt.qname,)):  # allow non-2.1 to be ok regardless per RH 2013-03-13
                    val.modelXbrl.error("SBR.NL.2.3.0.11",
                        _("Arc element %(element)s may not be contained in a linkbase with %(element2)s"),
                        modelObject=elt, element=elt.qname, element2=val.extendedElementName)
                if elt.qname == XbrlConst.qnLinkLabelArc and elt.get("order"):
                    val.modelXbrl.error("SBR.NL.2.3.8.08",
                        _("labelArc may not be contain order (%(order)s)"),
                        modelObject=elt, order=elt.get("order"))

            xmlLang = elt.get("{http://www.w3.org/XML/1998/namespace}lang")
            if val.validateXmlLang and xmlLang is not None:
                if not val.disclosureSystem.xmlLangPattern.match(xmlLang): # corrected merge of pre-plugin code per LOGIUS
                    val.modelXbrl.error("SBR.NL.2.3.8.01" if xmlLang.startswith('nl') else "SBR.NL.2.3.8.02" if xmlLang.startswith('en') else "arelle:langError",
                        _("Element %(element)s %(xlinkLabel)s has unauthorized xml:lang='%(lang)s'"),
                        modelObject=elt, element=elt.qname,
                        xlinkLabel=elt.get("{http://www.w3.org/1999/xlink}label"),
                        lang=elt.get("{http://www.w3.org/XML/1998/namespace}lang"),
                        messageCodes=("SBR.NL.2.3.8.01", "SBR.NL.2.3.8.02", "arelle:langError"))
            # check attributes for prefixes and xmlns
            val.valUsedPrefixes.add(elt.prefix)
            if elt.namespaceURI not in val.disclosureSystem.baseTaxonomyNamespaces:
                val.modelXbrl.error("SBR.NL.2.2.0.20",
                    _("%(fileType)s element %(element)s must not have custom namespace %(namespace)s"),
                    modelObject=elt, element=elt.qname,
                    fileType="schema" if isSchema else "linkbase" ,
                    namespace=elt.namespaceURI)
            for attrTag, attrValue in elt.items():
                prefix, ns, localName = XmlUtil.clarkNotationToPrefixNsLocalname(elt, attrTag, isAttribute=True)
                if prefix: # don't count unqualified prefixes for using default namespace
                    val.valUsedPrefixes.add(prefix)
                if ns and ns not in val.disclosureSystem.baseTaxonomyNamespaces:
                    val.modelXbrl.error("SBR.NL.2.2.0.20",
                        _("%(fileType)s element %(element)s must not have %(prefix)s:%(localName)s"),
                        modelObject=elt, element=elt.qname,
                        fileType="schema" if isSchema else "linkbase" ,
                        prefix=prefix, localName=localName)
                if isSchema and localName in ("base", "ref", "substitutionGroup", "type"):
                    valuePrefix, sep, valueName = attrValue.partition(":")
                    if sep:
                        val.valUsedPrefixes.add(valuePrefix)
            # check for xmlns on a non-root element
            parentElt = elt.getparent()
            if parentElt is not None:
                for prefix, ns in elt.nsmap.items():
                    if prefix not in parentElt.nsmap or parentElt.nsmap[prefix] != ns:
                        val.modelXbrl.error(("SBR.NL.2.2.0.19" if isSchema else "SBR.NL.2.3.1.01"),
                            _("%(fileType)s element %(element)s must not have xmlns:%(prefix)s"),
                            modelObject=elt, element=elt.qname,
                            fileType="schema" if isSchema else "linkbase" ,
                            prefix=prefix,
                            messageCodes=("SBR.NL.2.2.0.19", "SBR.NL.2.3.1.01"))

            if elt.localName == "roleType" and not elt.get("id"):
                val.modelXbrl.error("SBR.NL.2.3.10.11",
                    _("RoleType %(roleURI)s missing id attribute"),
                    modelObject=elt, roleURI=elt.get("roleURI"))
            elif elt.localName == "loc" and elt.get("{http://www.w3.org/1999/xlink}role"):
                val.modelXbrl.error("SBR.NL.2.3.10.08",
                    _("Loc %(xlinkLabel)s has unauthorized role attribute"),
                    modelObject=elt, xlinkLabel=elt.get("{http://www.w3.org/1999/xlink}label"))
            elif elt.localName == "documentation":
                val.modelXbrl.error("SBR.NL.2.3.10.12" if elt.namespaceURI == XbrlConst.link else "SBR.NL.2.2.11.02",
                    _("Documentation element must not be used: %(value)s"),
                    modelObject=elt, value=XmlUtil.text(elt),
                    messageCodes=("SBR.NL.2.3.10.12", "SBR.NL.2.2.11.02"))
            if elt.localName == "linkbase":
                schemaLocation = elt.get("{http://www.w3.org/2001/XMLSchema-instance}schemaLocation")
                if schemaLocation:
                    schemaLocations = schemaLocation.split()
                    for sl in (XbrlConst.link, XbrlConst.xlink):
                        if sl in schemaLocations:
                            val.modelXbrl.error("SBR.NL.2.3.0.07",
                                _("Linkbase element must not have schemaLocation entry for %(schemaLocation)s"),
                                modelObject=elt, schemaLocation=sl)
                for attrName, errCode in (("id", "SBR.NL.2.3.10.04"),
                                          ("{http://www.w3.org/2001/XMLSchema-instance}nil", "SBR.NL.2.3.10.05"),
                                          ("{http://www.w3.org/2001/XMLSchema-instance}noNamespaceSchemaLocation", "SBR.NL.2.3.10.06"),
                                          ("{http://www.w3.org/2001/XMLSchema-instance}type", "SBR.NL.2.3.10.07")):
                    if elt.get(attrName) is not None:
                        val.modelXbrl.error(errCode,
                            _("Linkbase element %(element)s must not have attribute %(attribute)s"),
                            modelObject=elt, element=elt.qname, attribute=attrName,
                            messageCodes=("SBR.NL.2.3.10.04", "SBR.NL.2.3.10.05", "SBR.NL.2.3.10.06", "SBR.NL.2.3.10.07"))
            for attrName, errCode in (("{http://www.w3.org/1999/xlink}actuate", "SBR.NL.2.3.10.01"),
                                      ("{http://www.w3.org/1999/xlink}show", "SBR.NL.2.3.10.02"),
                                      ("{http://www.w3.org/1999/xlink}title", "SBR.NL.2.3.10.03")):
                if elt.get(attrName) is not None:
                    val.modelXbrl.error(errCode,
                        _("Linkbase element %(element)s must not have attribute xlink:%(attribute)s"),
                        modelObject=elt, element=elt.qname, attribute=attrName,
                        messageCodes=("SBR.NL.2.3.10.01", "SBR.NL.2.3.10.02", "SBR.NL.2.3.10.03"))
        elif isinstance(elt,ModelComment): # comment node
            if elt.itersiblings(preceding=True): # corrected merge of pre-plugin code per LOGIUS
                val.modelXbrl.error("SBR.NL.2.2.0.05" if isSchema else "SBR.NL.2.3.0.05",
                        _('%(fileType)s must have only one comment node before schema element: "%(value)s"'),
                        modelObject=elt, fileType=modelDocument.gettype().title(), value=elt.text,
                        messageCodes=("SBR.NL.2.2.0.05", "SBR.NL.2.3.0.05"))

    # check folder names
    if modelDocument.filepathdir.startswith(modelXbrl.uriDir):
        partnerPrefix = None
        pathDir = modelDocument.filepathdir[len(modelXbrl.uriDir) + 1:].replace("\\", "/")
        lastPathSegment = None
        for pathSegment in pathDir.split("/"):
            if pathSegment.lower() != pathSegment:
                modelXbrl.error("SBR.NL.3.2.1.02",
                    _("Folder names must be in lower case: %(folder)s"),
                    modelObject=modelDocument, folder=pathSegment)
            if len(pathSegment) >= 15 :
                modelXbrl.error("SBR.NL.3.2.1.03",
                    _("Folder names must be less than 15 characters: %(folder)s"),
                    modelObject=modelDocument, folder=pathSegment)
            if pathSegment in ("bd", "kvk", "cbs"):
                partnerPrefix = pathSegment + '-'
            lastPathSegment = pathSegment
        if modelDocument.basename.lower() != modelDocument.basename:
            modelXbrl.error("SBR.NL.3.2.1.05",
                _("File names must be in lower case: %(file)s"),
                modelObject=modelDocument, file=modelDocument.basename)
        if partnerPrefix and not modelDocument.basename.startswith(partnerPrefix):
            modelXbrl.error("SBR.NL.3.2.1.14",
                "NT Partner DTS files MUST start with %(partnerPrefix)s consistently: %(filename)s",
                modelObject=modelDocument, partnerPrefix=partnerPrefix, filename=modelDocument.uri)
        if modelDocument.type == ModelDocument.Type.SCHEMA:
            if modelDocument.targetNamespace:
                nsParts = modelDocument.targetNamespace.split("/")
                # [0] = https, [1] = // [2] = nl.taxonomie  [3] = year or version
                nsYrOrVer = nsParts[3]
                requiredNamespace = "http://www.nltaxonomie.nl/" + nsYrOrVer + "/" + pathDir + "/" + modelDocument.basename[:-4]
                requiredLinkrole = "http://www.nltaxonomie.nl/" + nsYrOrVer + "/" + pathDir + "/"
                if modelDocument == modelXbrl.modelDocument:  # entry point
                    nsYr = "{year}"
                    if '2009' <= nsParts[3] < '2020':  # must be a year, use as year
                        nsYr = nsParts[3]
                    else: # look for year in parts of basename of required namespace
                        for nsPart in nsParts:
                            for baseNamePart in nsPart.split('-'):
                                if '2009' <= baseNamePart < '2020':
                                    nsYr = baseNamePart
                                    break
                    if not requiredNamespace.endswith('-' + nsYr):
                        requiredNamespace += '-' + nsYr
                if not modelDocument.targetNamespace.startswith(requiredNamespace):
                    modelXbrl.error("SBR.NL.3.2.3.06",
                        _("Namespace URI's MUST be constructed like %(requiredNamespace)s: %(namespaceURI)s"),
                        modelObject=modelDocument, requiredNamespace=requiredNamespace, namespaceURI=modelDocument.targetNamespace)
            else:
                requiredLinkrole = ''
            # concept checks
            for modelConcept in modelDocument.xmlRootElement.iterdescendants(tag="{http://www.w3.org/2001/XMLSchema}element"):
                if isinstance(modelConcept, ModelConcept):
                    # 6.7.16 name not duplicated in standard taxonomies
                    name = modelConcept.get("name")
                    if name:
                        ''' removed per RH 2013-03-25
                        substititutionGroupQname = modelConcept.substitutionGroupQname
                        if substititutionGroupQname:
                            if name.endswith("Member") ^ (substititutionGroupQname.localName == "domainMemberItem" and
                                                          substititutionGroupQname.namespaceURI.endswith("/xbrl/xbrl-syntax-extension")):
                                modelXbrl.error("SBR.NL.3.2.5.11",
                                    _("Concept %(concept)s must end in Member to be in sbr:domainMemberItem substitution group"),
                                    modelObject=modelConcept, concept=modelConcept.qname)
                            if name.endswith("Domain") ^ (substititutionGroupQname.localName == "domainItem" and
                                                          substititutionGroupQname.namespaceURI.endswith("/xbrl/xbrl-syntax-extension")):
                                modelXbrl.error("SBR.NL.3.2.5.12",
                                    _("Concept %(concept)s must end in Domain to be in sbr:domainItem substitution group"),
                                    modelObject=modelConcept, concept=modelConcept.qname)
                            if name.endswith("TypedAxis") ^ (substititutionGroupQname == XbrlConst.qnXbrldtDimensionItem and
                                                             modelConcept.isTypedDimension):
                                modelXbrl.error("SBR.NL.3.2.5.14",
                                    _("Concept %(concept)s must end in TypedAxis to be in xbrldt:dimensionItem substitution group if they represent a typed dimension"),
                                    modelObject=modelConcept, concept=modelConcept.qname)
                            if (name.endswith("Axis") and
                                not name.endswith("TypedAxis")) ^ (substititutionGroupQname == XbrlConst.qnXbrldtDimensionItem and
                                                                   modelConcept.isExplicitDimension):
                                modelXbrl.error("SBR.NL.3.2.5.13",
                                    _("Concept %(concept)s must end in Axis to be in xbrldt:dimensionItem substitution group if they represent an explicit dimension"),
                                    modelObject=modelConcept, concept=modelConcept.qname)
                            if name.endswith("Table") ^ (substititutionGroupQname == XbrlConst.qnXbrldtHypercubeItem):
                                modelXbrl.error("SBR.NL.3.2.5.15",
                                    _("Concept %(concept)s must end in Table to be in xbrldt:hypercubeItem substitution group"),
                                    modelObject=modelConcept, concept=modelConcept.qname)
                            if name.endswith("Title") ^ (substititutionGroupQname.localName == "presentationItem" and
                                                         substititutionGroupQname.namespaceURI.endswith("/xbrl/xbrl-syntax-extension")):
                                modelXbrl.error("SBR.NL.3.2.5.16",
                                    _("Concept %(concept)s must end in Title to be in sbr:presentationItem substitution group"),
                                    modelObject=modelConcept, concept=modelConcept.qname)
                        '''
                        if len(name) > 200:
                            modelXbrl.error("SBR.NL.3.2.12.02" if modelConcept.isLinkPart
                                                else "SBR.NL.3.2.5.21" if (modelConcept.isItem or modelConcept.isTuple)
                                                else "SBR.NL.3.2.14.01",
                                _("Concept %(concept)s name length %(namelength)s exceeds 200 characters"),
                                modelObject=modelConcept, concept=modelConcept.qname, namelength=len(name))
            # type checks
            for typeType in ("simpleType", "complexType"):
                for modelType in modelDocument.xmlRootElement.iterdescendants(tag="{http://www.w3.org/2001/XMLSchema}" + typeType):
                    if isinstance(modelType, ModelType):
                        name = modelType.get("name")
                        if name is None:
                            name = ""
                            if modelType.get("ref") is not None:
                                continue    # don't validate ref's here
                        if len(name) > 200:
                            modelXbrl.error("SBR.NL.3.2.5.21",
                                _("Type %(type)s name length %(namelength)s exceeds 200 characters"),
                                modelObject=modelType, type=modelType.qname, namelength=len(name))
                        if modelType.qnameDerivedFrom and modelType.qnameDerivedFrom.namespaceURI != XbrlConst.xbrli:
                            modelXbrl.error("SBR.NL.3.2.8.01",
                                _("Custom datatypes MUST be a restriction from XII defined datatypes: %(type)s"),
                                modelObject=modelType, type=modelType.qname)
                        if re.match(r"[^a-zA-Z0-9_-]", name):
                            modelXbrl.error("SBR.NL.3.2.8.02",
                                _("Datatype names MUST use characters a-zA-Z0-9_- only: %(type)s"),
                                modelObject=modelDocument, type=modelType.qname)
                        if modelType.facets and "enumeration" in modelType.facets:
                            if not modelType.qnameDerivedFrom == XbrlConst.qnXbrliStringItemType:
                                modelXbrl.error("SBR.NL.3.2.13.01",
                                    _("Enumerations MUST use a restriction on xbrli:stringItemType: %(type)s"),
                                    modelObject=modelDocument, type=modelType.qname)
            if lastPathSegment == "entrypoints":
                if not modelDocument.xmlRootElement.id:
                    modelXbrl.error("SBR.NL.2.2.0.23",
                        _("xs:schema/@id MUST be present in schema files in the reports/{NT partner}/entrypoints/ folder"),
                        modelObject=modelDocument)


    # check for idObject conflicts
    for modelObject in modelDocument.xmlRootElement.iter():
        if isinstance(modelObject,ModelObject):
            id = modelObject.id
            if id:
                if id in val.idObjects:
                    modelXbrl.error("SBR.NL.3.2.6.01",
                        _("ID %(id)s must be unique in the DTS but is present on two elements."),
                        modelObject=(modelObject, val.idObjects[id]), id=id)
                else:
                    val.idObjects[id] = modelObject


    for roleURI, modelRoleTypes in modelXbrl.roleTypes.items():
        if not roleURI.startswith("http://www.xbrl.org"):
            usedOns = set.union(*[modelRoleType.usedOns for modelRoleType in modelRoleTypes])
            # check roletypes for linkroles (only)
            if usedOns & {XbrlConst.qnLinkPresentationLink, XbrlConst.qnLinkCalculationLink, XbrlConst.qnLinkDefinitionLink,
                          XbrlConst.qnLinkLabel, XbrlConst.qnLinkReference, XbrlConst.qnLinkFootnote}:
                if len(modelRoleTypes) > 1:
                    modelXbrl.error("SBR.NL.3.2.9.01",
                        _("Linkrole URI's MUST be unique in the NT: %(linkrole)s"),
                        modelObject=modelRoleTypes, linkrole=roleURI)
                if roleURI.lower() != roleURI:
                    modelXbrl.error("SBR.NL.3.2.9.02",
                        _("Linkrole URI's MUST be in lowercase: %(linkrole)s"),
                        modelObject=modelRoleTypes, linkrole=roleURI)
                if re.match(r"[^a-z0-9_/-]", roleURI):
                    modelXbrl.error("SBR.NL.3.2.9.03",
                        _("Linkrole URI's MUST use characters a-z0-9_-/ only: %(linkrole)s"),
                        modelObject=modelRoleTypes, linkrole=roleURI)
                if len(roleURI) > 255:
                    modelXbrl.error("SBR.NL.3.2.9.04",
                        _("Linkrole URI's MUST NOT be longer than 255 characters, length is %(len)s: %(linkrole)s"),
                        modelObject=modelRoleTypes, len=len(roleURI), linkrole=roleURI)
                ''' removed per RH 2013-03-13 e-mail
                if not roleURI.startswith('http://www.nltaxonomie.nl'):
                    modelXbrl.error("SBR.NL.3.2.9.05",
                        _("Linkrole URI's MUST start with 'http://www.nltaxonomie.nl': %(linkrole)s"),
                        modelObject=modelRoleTypes, linkrole=roleURI)
                if (requiredLinkrole and
                    not roleURI.startswith(requiredLinkrole) and
                    re.match(r".*(domain$|axis$|table$|lineitem$)", roleURI)):
                        modelXbrl.error("SBR.NL.3.2.9.06",
                            _("Linkrole URI's MUST have the following construct: http://www.nltaxonomie.nl / {folder path} / {functional name} - {domain or axis or table or lineitem}: %(linkrole)s"),
                            modelObject=modelRoleTypes, linkrole=roleURI)
                '''
                for modelRoleType in modelRoleTypes:
                    if len(modelRoleType.id) > 255:
                        modelXbrl.error("SBR.NL.3.2.10.02",
                            _("Linkrole @id MUST NOT exceed 255 characters, length is %(length)s: %(linkroleID)s"),
                            modelObject=modelRoleType, length=len(modelRoleType.id), linkroleID=modelRoleType.id)
                partnerPrefix = modelRoleTypes[0].modelDocument.basename.split('-')
                if partnerPrefix:  # first element before dash is prefix
                    urnPartnerLinkroleStart = "urn:{0}:linkrole:".format(partnerPrefix[0])
                    if not roleURI.startswith(urnPartnerLinkroleStart):
                        modelXbrl.error("SBR.NL.3.2.9.10",
                            _("Linkrole MUST start with urn:{NT partner code}:linkrole:, \nexpecting: %(expectedStart)s..., \nfound: %(linkrole)s"),
                            modelObject=modelRoleType, expectedStart=urnPartnerLinkroleStart, linkrole=roleURI)

    if modelDocument.type in (ModelDocument.Type.SCHEMA, ModelDocument.Type.LINKBASE):
        # check namespaces are used
        for prefix, ns in modelDocument.xmlRootElement.nsmap.items():
            if ((prefix not in val.valUsedPrefixes) and
                (modelDocument.type != ModelDocument.Type.SCHEMA or ns != modelDocument.targetNamespace)):
                modelXbrl.error("SBR.NL.2.2.0.11" if modelDocument.type == ModelDocument.Type.SCHEMA else "SBR.NL.2.3.0.08",
                    _('%(docType)s namespace declaration "%(declaration)s" is not used'),
                    modelObject=modelDocument, docType=modelDocument.gettype().title(),
                    declaration=("xmlns" + (":" + prefix if prefix else "") + "=" + ns))

    del val.valUsedPrefixes



from __future__ import annotations
from typing import Any
import locale
import pytest
from decimal import Decimal
from arelle.Locale import format_decimal
from arelle.Locale import getUserLocale


d = Decimal('-1234567.8901')

@pytest.mark.parametrize(
    'params, result',
    [
        (
            {
                'conv': getUserLocale()[0],
                'value': d,
                'curr': '$',
                'neg': '-'
            },
            '-$1,234,567.89'
        ),
        (
            {
                'conv': getUserLocale()[0],
                'value': d,
                'fractPlaces': 0,
                'sep': '.',
                'dp': '',
                'trailneg': '-'
            },
            '1.234.568-'
        ),
        (
            {
                'conv': getUserLocale()[0],
                'value': d,
                'curr': '$',
                'neg': '(',
                'trailneg': ')'
            },
            '($1,234,567.89)'
        ),
        (
            {
                'conv': getUserLocale()[0],
                'value': Decimal(123456789),
                'sep': ' '
            },
            '123 456 789.00'
        ),
        (
            {
                'conv': getUserLocale()[0],
                'value': Decimal('-0.02'),
                'neg': '<',
                'trailneg': '>'
            },
            '<0.02>'
        ),
    ]
)
def test_format_decimal(params: dict[str, Any], result: str) -> None:
    assert format_decimal(**params) == result


@pytest.mark.parametrize('locale_code', ['', 'C', 'invalid'])
def test_get_user_locale_reset(locale_code) -> None:
    before_locale = locale.setlocale(locale.LC_ALL)
    getUserLocale(locale_code)
    after_locale = locale.setlocale(locale.LC_ALL)
    assert after_locale == before_locale



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import cast

from arelle.ModelInstanceObject import ModelFact, ModelContext
from arelle.ModelValue import QName
from arelle.ModelXbrl import ModelXbrl
from arelle.utils.PluginData import PluginData


@dataclass
class PluginValidationDataExtension(PluginData):
    annualReportTypes: frozenset[str]
    assetsQn: QName
    classOfReportingEntityQn: QName
    consolidatedMemberQn: QName
    consolidatedSoloDimensionQn: QName
    dateOfApprovalOfAnnualReportQn: QName
    dateOfExtraordinaryDividendDistributedAfterEndOfReportingPeriod: QName
    dateOfGeneralMeetingQn: QName
    descriptionOfQualificationsOfAssuranceEngagementPerformedQn: QName
    distributionOfResultsQns: frozenset[QName]
    equityQn: QName
    extraordinaryCostsQn: QName
    extraordinaryIncomeQn: QName
    extraordinaryResultBeforeTaxQn: QName
    fr37RestrictedText: str
    informationOnTypeOfSubmittedReportQn: QName
    liabilitiesQn: QName
    positiveProfitThreshold: float
    precedingReportingPeriodEndDateQn: QName
    precedingReportingPeriodStartDateQn: QName
    profitLossQn: QName
    proposedDividendRecognisedInEquityQn: QName
    reportingPeriodEndDateQn: QName
    reportingPeriodStartDateQn: QName
    taxExpenseOnOrdinaryActivitiesQn: QName
    taxExpenseQn: QName
    typeOfReportingPeriodDimensionQn: QName

    _contextFactMap: dict[str, dict[QName, ModelFact]] | None = None
    _reportingPeriodContexts: list[ModelContext] | None = None

    def contextFactMap(self, modelXbrl: ModelXbrl) -> dict[str, dict[QName, ModelFact]]:
        if self._contextFactMap is None:
            self._contextFactMap = defaultdict(dict)
            for fact in modelXbrl.facts:
                self._contextFactMap[fact.contextID][fact.qname] = fact
        return self._contextFactMap

    def getCurrentAndPreviousReportingPeriodContexts(self, modelXbrl: ModelXbrl) -> list[ModelContext]:
        """
        :return: Returns the most recent reporting period contexts (at most two).
        """
        contexts = self.getReportingPeriodContexts(modelXbrl)
        if not contexts:
            return contexts
        if len(contexts) > 2:
            return contexts[-2:]
        return contexts

    def getReportingPeriodContexts(self, modelXbrl: ModelXbrl) -> list[ModelContext]:
        """
        :return: A sorted list of contexts that match "reporting period" criteria.
        """
        if self._reportingPeriodContexts is not None:
            return self._reportingPeriodContexts
        contexts = []
        for context in modelXbrl.contexts.values():
            context = cast(ModelContext, context)
            if context.isInstantPeriod or context.isForeverPeriod:
                continue  # Reporting period contexts can't be instant/forever contexts
            if len(context.qnameDims) > 0:
                if context.qnameDims.keys() != {self.consolidatedSoloDimensionQn}:
                    continue  # Context is dimensionalized with something other than consolidatedSoloDimensionQn
                if context.dimMemberQname(self.consolidatedSoloDimensionQn) != self.consolidatedMemberQn:
                    continue  # Context is dimensionalized with the correct dimension but not member
            contexts.append(context)
        self._reportingPeriodContexts = sorted(contexts, key=lambda c: c.endDatetime)
        return self._reportingPeriodContexts



'''
See COPYRIGHT.md for copyright information.
'''
from __future__ import annotations

import datetime
from decimal import Decimal, InvalidOperation
from fractions import Fraction
from numbers import Number
from typing import Any, Callable, Iterable, MutableSequence, Sequence, TYPE_CHECKING, Type, Union, cast

from lxml import etree

from arelle import XbrlConst, XmlUtil
from arelle.ModelInstanceObject import ModelContext, ModelFact, ModelInlineFact, ModelUnit
from arelle.ModelObject import ModelAttribute, ModelObject
from arelle.ModelValue import (
    AnyURI,
    DATE,
    DATETIME,
    DATEUNION,
    DateTime,
    DayTimeDuration,
    QName,
    Time,
    TypeXValue,
    YearMonthDuration,
    anyURI,
    dateTime,
    gDay,
    gMonth,
    gMonthDay,
    gYear,
    gYearMonth,
    qname,
)
from arelle.ModelXbrl import ModelXbrl
from arelle.PluginManager import pluginClassMethods
from arelle.PrototypeDtsObject import PrototypeElementTree, PrototypeObject
from arelle.PythonUtil import STR_NUM_TYPES
from arelle.formula.FactAspectsCache import FactAspectsCache
from arelle.formula.XPathParser import (
    Expr,
    FormulaToken,
    OperationDef,
    ProgHeader,
    QNameDef,
    RangeDecl,
    VariableRef,
    exceptionErrorIndication,
)
from arelle.XmlValidateConst import UNKNOWN, VALID, VALID_NO_CONTENT
from arelle.XmlValidate import validate as xmlValidate
from arelle.typing import TypeGetText

if TYPE_CHECKING:
    from arelle.ModelDocument import ModelDocument
    from arelle.ModelFormulaObject import FormulaOptions, ModelGeneral, Trace as TraceClass

_: TypeGetText

ContextItem = Union[
    DayTimeDuration,
    Decimal,
    Fraction,
    ModelAttribute,
    ModelObject,
    ModelXbrl,
    QName,
    Time,
    YearMonthDuration,
    bool,
    datetime.datetime,
    etree._ElementTree,
    float,
    int,
    range,
    str,
]

AtomizedValue = Union[
    TypeXValue,
    bool,
    int,
    range,
]

RecursiveContextItem = Union[ContextItem, Iterable['RecursiveContextItem']]
ResultStack = MutableSequence[Sequence[ContextItem]]

# deferred types initialization
boolean: Callable[[
    XPathContext,
    FormulaToken | None,
    ContextItem | None,
    ResultStack,
], bool] | None = None
testTypeCompatibility: Callable[[
    XPathContext,
    FormulaToken,
    str,
    ContextItem,
    ContextItem,
], None] | None = None
Trace: Type[TraceClass] | None = None
qnWild = qname("*")  # "*"


class XPathException(Exception):
    def __init__(self, progStep: FormulaToken | None, code: QName | str, message: str) -> None:
        self.column = None
        if isinstance(progStep, OperationDef):
            self.line = progStep.sourceStr
            self.column = progStep.loc
        elif isinstance(progStep, ProgHeader):
            self.line = progStep.sourceStr
        elif isinstance(progStep, XPathContext) and progStep.progHeader:
            self.line = progStep.progHeader.sourceStr
        else:
            self.line = "(not available)"
        self.code = str(code)  # called with qname or string, qname -> prefixed name string
        self.message = message
        self.args = (self.__repr__(),)

    def __repr__(self) -> str:
        if self.column:
            return _('[{0}] exception at {1} in {2}').format(self.code, self.column, self.message)
        else:
            return _('[{0}] exception {1}').format(self.code, self.message)

    @property
    def sourceErrorIndication(self) -> str:
        return exceptionErrorIndication(self)


class FunctionNumArgs(Exception):
    def __init__(self, errCode: str = 'err:XPST0017', errText: str | None = None) -> None:
        self.errCode = errCode
        self.errText = errText or _('Number of arguments do not match signature arity')
        self.args = (self.__repr__(),)

    def __repr__(self) -> str:
        return _("Exception: Number of arguments mismatch")


class FunctionArgType(Exception):
    def __init__(
            self,
            argIndex: int | str,
            expectedType: str,
            foundObject: str | QName | Sequence[FormulaToken] | None = '',
            errCode: str = 'err:XPTY0004',
            value: str | None = None
    ) -> None:
        self.errCode = errCode
        self.argNum = (argIndex + 1) if isinstance(argIndex, int) else argIndex
        self.expectedType = expectedType
        self.foundObject = foundObject
        self.value = value
        self.args = (self.__repr__(),)

    def __repr__(self) -> str:
        return _("[{0}]: Arg {1} expected type {2}").format(self.errCode, self.argNum, self.expectedType)


class FunctionNotAvailable(Exception):
    def __init__(self, name: str | None = None) -> None:
        self.name = name
        self.args = (self.__repr__(),)

    def __repr__(self) -> str:
        return _("Exception, function implementation not available: {0}").format(self.name)


class RunTimeExceededException(Exception):
    def __init__(self) -> None:
        self.args = (self.__repr__(),)

    def __repr__(self) -> str:
        return _("Formula run time exceeded")


def create(
        modelXbrl: ModelXbrl,
        inputXbrlInstance: ModelDocument | None = None,
        sourceElement: ModelObject | None = None,
) -> XPathContext:
    global boolean, testTypeCompatibility, Trace
    if boolean is None:
        from arelle.FunctionUtil import testTypeCompatibility
        from arelle.ModelFormulaObject import Trace
        from arelle.FunctionFn import boolean

    if inputXbrlInstance is None:
        inputXbrlInstance = modelXbrl.modelDocument
    assert inputXbrlInstance is not None
    return XPathContext(modelXbrl, inputXbrlInstance, sourceElement)


# note: 2.2% execution time savings by having these sets/lists as constant instead of in expression where used
VALUE_OPS = {'+', '-', '*', 'div', 'idiv', 'mod', 'to', 'gt', 'ge', 'eq', 'ne', 'lt', 'le'}
GENERALCOMPARISON_OPS = {'>', '>=', '=', '!=', '<', '<='}
NODECOMPARISON_OPS = {'is', '>>', '<<'}
COMBINING_OPS = {'intersect', 'except', 'union', '|'}
LOGICAL_OPS = {'and', 'or'}
UNARY_OPS = {'u+', 'u-'}
FORSOMEEVERY_OPS = {'for', 'some', 'every'}
PATH_OPS = {'/', '//', 'rootChild', 'rootDescendant'}
SEQUENCE_TYPES = (tuple, list, set)
GREGORIAN_TYPES = (gYearMonth, gYear, gMonthDay, gDay, gMonth)


class XPathContext:
    def __init__(
            self,
            modelXbrl: ModelXbrl,
            inputXbrlInstance: ModelDocument,
            sourceElement: ModelObject | None,
            inScopeVars: dict[QName, ModelXbrl | ModelObject | int | str] | None = None,
            factAspectsCache: FactAspectsCache | None = None,
    ) -> None:
        self.modelXbrl = modelXbrl
        self.isRunTimeExceeded = False
        self.inputXbrlInstance = inputXbrlInstance
        self.outputLastContext: dict[QName, ModelContext] = {}  # last context element output per output instance
        self.outputLastUnit: dict[QName, ModelUnit] = {}
        self.outputLastFact: dict[QName, ModelFact] = {}
        self.outputFirstFact: dict[QName, ModelFact] = {}
        self.sourceElement: ModelObject | None = sourceElement
        self.contextItem = self.inputXbrlInstance.targetXbrlRootElement
        self.progHeader: ProgHeader | None = None
        self.traceType: int | None = None
        self.variableSet = None
        self.factAspectsCache = factAspectsCache or FactAspectsCache(modelXbrl.modelManager.formulaOptions.cacheSize)
        self.inScopeVars: dict[QName, ModelXbrl | ModelObject | int | str] = {} if inScopeVars is None else inScopeVars
        self.cachedFilterResults: dict[ModelGeneral, set[ModelFact]] = {}
        if inputXbrlInstance:
            self.inScopeVars[XbrlConst.qnStandardInputInstance] = inputXbrlInstance.modelXbrl
        self.customFunctions: dict[
            QName,
            Callable[[XPathContext, OperationDef, ContextItem, ResultStack], ContextItem]
        ] = {}
        for pluginXbrlMethod in pluginClassMethods("Formula.CustomFunctions"):
            self.customFunctions.update(pluginXbrlMethod())

    def copy(self) -> XPathContext:  # shallow copy (for such as for Table LB table processiong
        xpCtxCpy = XPathContext(self.modelXbrl, self.inputXbrlInstance, self.sourceElement, self.inScopeVars.copy())
        # note: not currently duplicating cachedFilterResults
        return xpCtxCpy

    def close(self) -> None:
        self.factAspectsCache.clear()
        self.outputLastContext.clear()  # dereference
        self.outputLastUnit.clear()
        self.outputLastFact.clear()
        self.outputFirstFact.clear()
        self.inScopeVars.clear()
        self.cachedFilterResults.clear()
        self.__dict__.clear()  # dereference everything

    def runTimeExceededCallback(self) -> None:
        self.isRunTimeExceeded = True

    @property
    def formulaOptions(self) -> FormulaOptions:
        return self.modelXbrl.modelManager.formulaOptions

    def evaluate(
            self,
            exprStack: Iterable[FormulaToken],
            contextItem: ContextItem | None = None,
            resultStack: ResultStack | None = None,
            parentOp: str | None = None,
    ) -> ResultStack:
        if resultStack is None:
            resultStack = []
        if contextItem is None:
            contextItem = self.contextItem
        setProgHeader = False
        for p in exprStack:
            result: RecursiveContextItem | None = None
            if isinstance(p, QNameDef) or (p == '*' and parentOp in ('/', '//')):  # path step QName or wildcard
                # step axis operation
                if len(resultStack) == 0 or not self.isNodeSequence(resultStack[-1]):
                    resultStack.append([contextItem, ])
                result = self.stepAxis(parentOp, p, resultStack.pop())
            elif isinstance(p, STR_NUM_TYPES):
                result = p
            elif isinstance(p, VariableRef):
                if p.name in self.inScopeVars:
                    result = self.inScopeVars[p.name]
                    # uncomment to allow lambdas as variable values (for deferred processing if needed)
                    # if isinstance(result, LambdaType):
                    #    result = result()  # dereference lambda-valued variables
                    if result is None:  # None atomic result is XPath empty sequence
                        result = []  # subsequent processing discards None results
            elif isinstance(p, OperationDef):
                op = p.name
                if isinstance(op, QNameDef):  # function call
                    args = self.evaluate(p.args, contextItem=contextItem)
                    ns = op.namespaceURI
                    localname = op.localName
                    try:
                        from arelle import FunctionXs, FunctionFn, FunctionXfi, FunctionIxt, FunctionCustom

                        if op in self.modelXbrl.modelCustomFunctionSignatures:
                            result = FunctionCustom.call(self, p, op, contextItem, args)
                        elif op in self.customFunctions:  # plug in method custom functions
                            result = self.customFunctions[op](self, p, contextItem, args)  # use plug-in's method
                        elif op.unprefixed and localname in {
                            'attribute',
                            'comment',
                            'document-node',
                            'element',
                            'item',
                            'node',
                            'processing-instruction',
                            'schema-attribute',
                            'schema-element',
                            'text',
                        }:
                            # step axis operation
                            if len(resultStack) == 0 or not self.isNodeSequence(resultStack[-1]):
                                if isinstance(contextItem, (tuple, list)):
                                    resultStack.append(contextItem)
                                else:
                                    resultStack.append([contextItem, ])
                            result = self.stepAxis(parentOp, p, resultStack.pop())
                        elif op.unprefixed or ns == XbrlConst.fn:
                            result = FunctionFn.call(self, p, localname, contextItem, args)
                        elif ns == XbrlConst.xfi or ns == XbrlConst.xff:
                            result = FunctionXfi.call(self, p, localname, args)
                        elif ns == XbrlConst.xsd:
                            result = FunctionXs.call(self, p, localname, args)
                        elif ns in FunctionIxt.ixtNamespaceFunctions:
                            result = FunctionIxt.call(self, p, op, args)
                        elif (
                                self.modelXbrl.modelManager.customTransforms is not None
                                and op in self.modelXbrl.modelManager.customTransforms
                        ):
                            result = self.modelXbrl.modelManager.customTransforms[op](cast(str, args[0][0]))
                        else:
                            raise XPathException(p, 'err:XPST0017', _('Function call not identified: {0}.').format(op))
                    except FunctionNumArgs as err:
                        raise XPathException(p, err.errCode, "{}: {}".format(err.errText, op))
                    except FunctionArgType as err:
                        raise XPathException(p, err.errCode, _('Argument {0} does not match expected type {1} for {2} {3}.').format(
                            err.argNum, err.expectedType, op, err.foundObject))
                    except FunctionNotAvailable:
                        raise XPathException(p, 'err:XPST0017', _('Function named {0} does not have a custom or built-in implementation.').format(op))
                elif op in VALUE_OPS:
                    # binary arithmetic operations and value comparisons
                    s1 = self.atomize(p, resultStack.pop()) if len(resultStack) > 0 else []
                    s2 = self.atomize(p, self.evaluate(p.args, contextItem=contextItem))
                    # value comparisons
                    if len(s1) > 1 or len(s2) > 1:
                        raise XPathException(p, 'err:XPTY0004', _("Value operation '{0}' sequence length error").format(op))
                    if len(s1) == 0 or len(s2) == 0:
                        result = []
                    else:
                        op1 = s1[0]
                        op2 = s2[0]
                        assert testTypeCompatibility is not None
                        testTypeCompatibility(self, p, op, op1, op2)
                        if type(op1) != type(op2) and op in (
                            '+',
                            '-',
                            '*',
                            'div',
                            'idiv',
                            'mod',
                            'ge',
                            'gt',
                            'le',
                            'lt',
                            'eq',
                            'ne',
                        ):
                            # check if type promotion needed (Decimal-float, not needed for integer-Decimal)
                            if isinstance(op1, Decimal) and isinstance(op2, float):
                                op1 = float(op1)  # per http://http://www.w3.org/TR/xpath20/#dt-type-promotion 1b
                            elif isinstance(op2, Decimal) and isinstance(op1, float):
                                op2 = float(op2)
                        if op == '+':
                            result = op1 + op2
                        elif op == '-':
                            result = op1 - op2
                        elif op == '*':
                            result = op1 * op2
                        elif op in ('div', 'idiv', "mod"):
                            try:
                                if op == 'div':
                                    result = op1 / op2
                                elif op == 'idiv':
                                    result = op1 // op2
                                elif op == 'mod':
                                    result = op1 % op2
                            except ZeroDivisionError:
                                raise XPathException(p, 'err:FOAR0001', _('Attempt to divide by zero: {0} {1} {2}.').format(op1, op, op2))
                        elif op == 'ge':
                            result = op1 >= op2
                        elif op == 'gt':
                            result = op1 > op2
                        elif op == 'le':
                            result = op1 <= op2
                        elif op == 'lt':
                            result = op1 < op2
                        elif op == 'eq':
                            result = op1 == op2
                        elif op == 'ne':
                            result = op1 != op2
                        elif op == 'to':
                            result = range(int(op1), int(op2) + 1)
                elif op in GENERALCOMPARISON_OPS:
                    # general comparisons
                    s1 = self.atomize(p, resultStack.pop()) if len(resultStack) > 0 else []
                    s2 = self.atomize(p, self.evaluate(p.args, contextItem=contextItem))
                    result = []
                    assert testTypeCompatibility is not None
                    for op1 in s1:
                        for op2 in s2:
                            testTypeCompatibility(self, p, op, op1, op2)
                            if op == '>=':
                                result = op1 >= op2
                            elif op == '>':
                                result = op1 > op2
                            elif op == '<=':
                                result = op1 <= op2
                            elif op == '<':
                                result = op1 < op2
                            elif op == '=':
                                result = op1 == op2
                            elif op == '!=':
                                result = op1 != op2
                            if result:
                                break
                        if result:
                            break
                elif op in NODECOMPARISON_OPS:
                    # node comparisons
                    s1 = resultStack.pop() if len(resultStack) > 0 else []
                    s2 = self.evaluate(p.args, contextItem=contextItem)
                    if len(s1) > 1 or len(s2) > 1 or not self.isNodeSequence(s1) or not self.isNodeSequence(s2[0]):
                        raise XPathException(p, 'err:XPTY0004', _('Node comparison sequence error'))
                    if len(s1) == 0 or len(s2[0]) == 0:
                        result = []
                    else:
                        n1 = s1[0]
                        n2 = s2[0][0]
                        result = False
                        for op1 in s1:
                            for op2 in s2:
                                if op == 'is':
                                    result = n1 == n2
                                elif op == '>>':
                                    result = op1 > op2
                                elif op == '<<':
                                    result = op1 <= op2
                            if result:
                                break
                elif op in COMBINING_OPS:
                    # node comparisons
                    s1 = resultStack.pop() if len(resultStack) > 0 else []
                    s2 = self.flattenSequence(self.evaluate(p.args, contextItem=contextItem))
                    if not self.isNodeSequence(s1) or not self.isNodeSequence(s2):
                        raise XPathException(p, 'err:XPTY0004', _('Node operation sequence error'))
                    set1 = set(s1)
                    set2 = set(s2)
                    if op == 'intersect':
                        resultset = set1 & set2
                    elif op == 'except':
                        resultset = set1 - set2
                    elif op == 'union' or op == '|':
                        resultset = set1 | set2
                    # convert to a list in document order
                    result = self.documentOrderedNodes(resultset)
                elif op in LOGICAL_OPS:
                    # general comparisons
                    if len(resultStack) == 0:
                        result = []
                    else:
                        op1 = self.effectiveBooleanValue(p, resultStack.pop()) if len(resultStack) > 0 else False
                        # consider short circuit possibilities
                        if op == 'or' and op1:
                            result = True
                        elif op == 'and' and not op1:
                            result = False
                        else:  # must evaluate other operand
                            op2 = self.effectiveBooleanValue(p, self.evaluate(p.args, contextItem=contextItem))
                            if op == 'and':
                                result = op1 and op2
                            elif op == 'or':
                                result = op1 or op2
                elif op in UNARY_OPS:
                    s1 = self.atomize(p, self.evaluate(p.args, contextItem=contextItem))
                    if len(s1) > 1:
                        raise XPathException(p, 'err:XPTY0004', _('Unary expression sequence length error'))
                    if len(s1) == 0:
                        result = []
                    else:
                        op1 = s1[0]
                        if op == 'u+':
                            result = op1
                        elif op == 'u-':
                            result = -op1
                elif op == 'instance':
                    result = False
                    s1 = self.flattenSequence(resultStack.pop()) if len(resultStack) > 0 else []
                    arity = len(s1)
                    if len(p.args) > 1:
                        occurenceIndicator = p.args[1]
                        if (
                            (occurenceIndicator == '?' and arity in (0, 1))
                            or (occurenceIndicator == '+' and arity >= 1)
                            or (occurenceIndicator == '*')
                        ):
                            result = True
                    elif arity == 1:
                        result = True
                    if result and len(p.args) > 0:
                        t = p.args[0]
                        for x in s1:
                            if isinstance(t, QNameDef):
                                if t.namespaceURI == XbrlConst.xsd:
                                    tType = {
                                        "integer": int,
                                        "string": str,
                                        "decimal": Decimal,
                                        "double": float,
                                        "float": float,
                                        "boolean": bool,
                                        "QName": QName,
                                        "anyURI": AnyURI,
                                        "date": DateTime,
                                        "dateTime": DateTime,
                                    }.get(t.localName)
                                    if tType:
                                        result = isinstance(x, tType)
                                        if result and tType == DateTime:
                                            result = x.dateOnly == (t.localName == "date")
                            elif isinstance(t, OperationDef):
                                if t.name == "element":
                                    if isinstance(x, ModelObject):
                                        if len(t.args) >= 1:
                                            qn = t.args[0]
                                            if qn == '*' or (isinstance(qn, QNameDef) and (qn == x.qname or qn == qnWild)):
                                                result = True
                                                if len(t.args) >= 2 and isinstance(t.args[1], QNameDef):
                                                    modelXbrl = x.modelDocument.modelXbrl
                                                    modelConcept = modelXbrl.qnameConcepts.get(x.qname)
                                                    if not modelConcept.instanceOfType(t.args[1]):
                                                        result = False
                                            else:
                                                result = False
                                    else:
                                        result = False
                                # elif t.name == "item" comes here and result stays True
                            if not result:
                                break
                elif op == 'sequence':
                    result = self.evaluate(p.args, contextItem=contextItem)
                elif op == 'predicate':
                    result = self.predicate(p, resultStack.pop()) if len(resultStack) > 0 else []
                elif op in FORSOMEEVERY_OPS:  # for, some, every
                    _result: ResultStack = []
                    self.evaluateRangeVars(op, p.args[0], p.args[1:], contextItem, _result)
                    result = _result
                elif op == 'if':
                    exprArg = cast(Expr, p.args[0])
                    exprArgStack = cast(Iterable[FormulaToken], exprArg.expr[0])
                    test = self.effectiveBooleanValue(p, self.evaluate(exprArgStack, contextItem=contextItem))
                    opDef = cast(OperationDef, p.args[1 if test else 2])
                    result = self.evaluate(opDef.args, contextItem=contextItem)
                elif op == '.':
                    result = contextItem
                elif op == '..':
                    result = XmlUtil.parent(cast(ModelObject, contextItem))
                elif op in PATH_OPS:
                    if op in ('rootChild', 'rootDescendant'):
                        # fix up for multi-instance
                        resultStack.append([self.inputXbrlInstance.targetXbrlElementTree, ])
                        op = '/' if op == 'rootChild' else '//'
                    # contains QNameDefs and predicates
                    innerFocusNodes: Sequence[ContextItem] | ContextItem
                    if len(resultStack) > 0:
                        innerFocusNodes = resultStack.pop()
                    else:
                        innerFocusNodes = contextItem
                    navSequence: ResultStack = []
                    innerFocusNode: ContextItem
                    for innerFocusNode in self.flattenSequence(innerFocusNodes):
                        navSequence += self.evaluate(p.args, contextItem=innerFocusNode, parentOp=op)
                    result = self.documentOrderedNodes(cast(Iterable[ContextItem], self.flattenSequence(navSequence)))
            elif isinstance(p, ProgHeader):
                self.progHeader = p
                assert Trace is not None
                if p.traceType not in (Trace.MESSAGE, Trace.CUSTOM_FUNCTION):
                    self.traceType = p.traceType
                setProgHeader = True
            if result is not None:  # note: result can be False which gets appended to resultStack
                resultStack.append(self.flattenSequence(result))
        if setProgHeader:
            self.progHeader = None
        return resultStack

    def evaluateBooleanValue(self, exprStack: Sequence[FormulaToken], contextItem: ContextItem | None = None) -> bool:
        if len(exprStack) > 0 and isinstance(exprStack[0], ProgHeader):
            progHeader = exprStack[0]
            return self.effectiveBooleanValue(progHeader, self.evaluate(exprStack, contextItem))
        return False

    def evaluateAtomicValue(
            self,
            exprStack: Sequence[FormulaToken],
            _type: QName | str | None,
            contextItem: ContextItem | None = None,
            resultMayBeNode: bool = False,
    ) -> Any:
        if exprStack and len(exprStack) > 0 and isinstance(exprStack[0], ProgHeader):
            progHeader = exprStack[0]
            eval = self.evaluate(exprStack, contextItem=contextItem)
            # TLB test case 200 v02i requires a node result (should xs:element be allowed?)
            if not _type and resultMayBeNode:
                result = self.flattenSequence(eval)
                if len(result) > 0 and self.isNodeSequence(result):
                    return result[0]
            result = self.atomize(progHeader, eval)
            if isinstance(_type, QName) and _type.namespaceURI == XbrlConst.xsd:
                _type = "xs:" + _type.localName
            if isinstance(_type, str):
                prefix, sep, localName = _type.rpartition(':')
                if prefix == 'xs':
                    if localName.endswith('*'):
                        localName = localName[:-1]
                    if isinstance(result, (tuple, list, set)):
                        from arelle import FunctionXs

                        if _type.endswith('*'):
                            return [FunctionXs.call(self, progHeader, localName, [r, ]) for r in result]
                        elif len(result) > 0:
                            return FunctionXs.call(self, progHeader, localName, [cast(Sequence[str], result)[0], ])
                elif localName.startswith("item()"):
                    return result  # can be any type
            else:  # no conversion
                if len(result) == 0:
                    return None
                elif len(result) == 1:
                    return result[0]
                else:
                    return result
        return None

    def evaluateRangeVars(
            self,
            op: str,
            p: FormulaToken,
            args: Sequence[FormulaToken],
            contextItem: ContextItem,
            result: Any,
    ) -> None:
        if isinstance(p, RangeDecl):
            evaluatedRangeDecl = self.evaluate(p.bindingSeq, contextItem=contextItem)
            if len(evaluatedRangeDecl) == 1:  # should be an expr single
                r = evaluatedRangeDecl[0]
                if isinstance(r, (tuple, list, set)):
                    if len(r) == 1 and isinstance(r[0], range):
                        r = r[0]
                    rvQname = p.rangeVar.name
                    hasPrevValue = rvQname in self.inScopeVars
                    if hasPrevValue:
                        prevValue = self.inScopeVars[rvQname]
                    for rv in r:
                        self.inScopeVars[rvQname] = rv  # type: ignore[assignment]
                        self.evaluateRangeVars(op, args[0], args[1:], contextItem, result)
                        if op != 'for' and len(result) > 0:
                            break  # short circuit evaluation
                    if op == 'every' and len(result) == 0:
                        result.append(True)  # true if no false result returned during iteration
                    if hasPrevValue:
                        self.inScopeVars[rvQname] = prevValue
        elif isinstance(p, Expr):
            if p.name == 'return':
                result.append(self.evaluate(cast(Iterable[FormulaToken], p.expr), contextItem=contextItem))
            elif p.name == 'satisfies':
                boolresult = self.effectiveBooleanValue(p, self.evaluate(cast(Iterable[FormulaToken], p.expr), contextItem=contextItem))
                if (op == 'every') != boolresult:
                    # stop short circuit eval
                    result.append(boolresult)

    def isNodeSequence(self, x: Iterable[ContextItem]) -> bool:
        for el in x:
            if not isinstance(el, ModelObject):
                return False
        return True

    def stepAxis(
            self,
            op: str | None,
            p: FormulaToken,
            sourceSequence: Iterable[ContextItem],
    ) -> Sequence[str | ModelAttribute | ModelObject]:
        targetSequence: list[str | ModelObject | ModelAttribute] = []
        for node in sourceSequence:
            if not isinstance(node, (ModelObject, etree._ElementTree, PrototypeElementTree, PrototypeObject, ModelAttribute)):
                raise XPathException(self.progHeader, 'err:XPTY0020', _('Axis step {0} context item is not a node: {1}').format(op, node))
            targetNodes: MutableSequence[str | ModelObject | ModelAttribute] = []
            if isinstance(p, QNameDef):
                ns = p.namespaceURI
                localname = p.localName
                axis = p.axis
                if p.isAttribute:
                    if isinstance(node, ModelObject):
                        attrTag = p.localName if p.unprefixed else p.clarkNotation
                        modelAttribute = None
                        try:
                            modelAttribute = node.xAttributes[attrTag]
                        except (AttributeError, TypeError, IndexError, KeyError):
                            # may be lax or deferred validated
                            try:
                                xmlValidate(node.modelXbrl, node, attrQname=p)
                                modelAttribute = node.xAttributes[attrTag]
                            except (AttributeError, TypeError, IndexError, KeyError):
                                pass
                        if modelAttribute is None:
                            value = node.get(attrTag)
                            if value is not None:
                                targetNodes.append(ModelAttribute(node, p.clarkNotation, UNKNOWN, value, value, value))
                        elif modelAttribute.xValid >= VALID or modelAttribute.xValid == UNKNOWN:  # may be undeclared attribute
                            targetNodes.append(modelAttribute)
                elif op == '/' or op is None:
                    if axis is None or axis == "child":
                        if isinstance(node, (ModelObject, etree._ElementTree, PrototypeElementTree, PrototypeObject)):
                            targetNodes = XmlUtil.children(cast(ModelObject, node), ns, localname, ixTarget=True)  # type: ignore[assignment]
                    elif axis == "parent":
                        parentNode: list[ModelObject | None]
                        if isinstance(node, ModelAttribute):
                            parentNode = [node.modelElement]
                        else:
                            parentNode = [XmlUtil.parent(cast(ModelObject, node))]
                        if (
                            isinstance(node, ModelObject)
                            and (not ns or ns == parentNode.namespaceURI or ns == "*")  # type: ignore[attr-defined]
                            and (localname == parentNode.localName or localname == "*")  # type: ignore[attr-defined]
                        ):
                            targetNodes = [parentNode]  # type: ignore[list-item]
                    elif axis == "self":
                        if (
                            isinstance(node, (ModelObject, PrototypeObject))
                            and (not ns or ns == node.namespaceURI or ns == "*")
                            and (localname == node.localName or localname == "*")
                        ):
                            targetNodes = [node]
                    elif axis.startswith("descendant"):
                        if isinstance(node, (ModelObject, etree._ElementTree, PrototypeElementTree, PrototypeObject)):
                            targetNodes = XmlUtil.descendants(node, ns, localname)  # type: ignore[assignment]
                            if (
                                axis.endswith("-or-self")
                                and isinstance(node, ModelObject)
                                and (not ns or ns == node.namespaceURI or ns == "*")
                                and (localname == node.localName or localname == "*")
                            ):
                                targetNodes.append(node)
                    elif axis.startswith("ancestor"):
                        if isinstance(node, (ModelObject, PrototypeObject)):
                            targetNodes = [
                                ancestor
                                for ancestor in XmlUtil.ancestors(node)
                                if (
                                    (not ns or ns == ancestor.namespaceURI or ns == "*")
                                    and (localname == ancestor.localName or localname == "*")
                                )
                            ]
                            if (
                                axis.endswith("-or-self")
                                and isinstance(node, ModelObject)
                                and (not ns or ns == node.namespaceURI or ns == "*")
                                and (localname == node.localName or localname == "*")
                            ):
                                targetNodes.insert(0, node)
                    elif axis.endswith("-sibling"):
                        if isinstance(node, ModelObject):
                            targetNodes = [
                                sibling  # type: ignore[misc]
                                for sibling in node.itersiblings(preceding=axis.startswith("preceding"))
                                if (
                                    (not ns or ns == sibling.namespaceURI or ns == "*")  # type: ignore[attr-defined]
                                    and (localname == sibling.localName or localname == "*")  # type: ignore[attr-defined]
                                )
                            ]
                    elif axis == "preceding":
                        if isinstance(node, ModelObject):
                            for preceding in cast(Iterable[ModelObject], node.getroottree().iter()):
                                if preceding == node:
                                    break
                                elif (
                                        (not ns or ns == preceding.namespaceURI or ns == "*")
                                        and (localname == preceding.localName or localname == "*")
                                ):
                                    targetNodes.append(preceding)
                    elif axis == "following" and isinstance(node, ModelObject):
                        foundNode = False
                        for following in cast(Iterable[ModelObject], node.getroottree().iter()):
                            if following == node:
                                foundNode = True
                            elif (
                                foundNode
                                and (not ns or ns == following.namespaceURI or ns == "*")
                                and (localname == following.localName or localname == "*")
                            ):
                                targetNodes.append(following)
                elif op == '//':
                    if isinstance(node, (ModelObject, etree._ElementTree, PrototypeElementTree, PrototypeObject)):
                        targetNodes = XmlUtil.descendants(node, ns, localname, ixTarget=True)  # type: ignore[assignment]
                elif op == '..':
                    if isinstance(node, ModelAttribute):
                        targetNodes = [node.modelElement]
                    else:
                        targetNodes = [XmlUtil.parent(cast(ModelObject, node), ixTarget=True)]  # type: ignore[list-item]
            elif isinstance(p, OperationDef) and isinstance(p.name, QNameDef):
                if isinstance(node, ModelObject) and p.name.localName == "text":
                    # note this is not string value, just child text
                    targetNodes = [node.textValue]
                    # todo: add element, attribute, node, etc...
            elif p == '*':  # wildcard
                if op == '/' or op is None:
                    if isinstance(node, (ModelObject, etree._ElementTree, PrototypeElementTree)):
                        targetNodes = XmlUtil.children(node, '*', '*', ixTarget=True)  # type: ignore[assignment]
                elif op == '//':
                    if isinstance(node, (ModelObject, etree._ElementTree, PrototypeElementTree)):
                        targetNodes = XmlUtil.descendants(node, '*', '*', ixTarget=True)  # type: ignore[assignment]
            targetSequence.extend(targetNodes)
        return targetSequence

    def predicate(self, p: OperationDef, sourceSequence: Iterable[ContextItem]) -> Sequence[ContextItem]:
        targetSequence = []
        sourcePosition = 0
        for item in sourceSequence:
            sourcePosition += 1
            predicateResult = self.evaluate(p.args, contextItem=item)
            if len(predicateResult) == 1:
                predicateResult = predicateResult[0]  # type: ignore[assignment]
            if len(predicateResult) == 1 and isinstance(predicateResult[0], Number):
                result = predicateResult[0]
                if isinstance(result, bool):  # note that bool is subclass of int
                    if result:
                        targetSequence.append(item)
                elif sourcePosition == result:  # type: ignore[comparison-overlap]
                    targetSequence.append(item)
            elif self.effectiveBooleanValue(p, predicateResult):
                targetSequence.append(item)
        return targetSequence

    def atomize(self, p: FormulaToken | None, x: RecursiveContextItem | None) -> Any:
        # sequence
        if isinstance(x, SEQUENCE_TYPES):
            sequence = []
            for item in self.flattenSequence(x):
                atomizedItem = self.atomize(p, item)
                if atomizedItem != []:
                    sequence.append(atomizedItem)
            return sequence
        # individual items
        if isinstance(x, range):
            return x
        baseXsdType = None
        e: ModelObject | None = None
        if isinstance(x, ModelFact):
            if x.isTuple:
                raise XPathException(p, 'err:FOTY0012', _('Atomizing tuple {0} that does not have a typed value').format(x))
            if x.isNil or x.concept is None:
                return []
            baseXsdType = x.concept.baseXsdType
            v = x.value  # resolves default value
            e = x
        elif isinstance(x, ModelAttribute):  # ModelAttribute is a tuple (below), check this first!
            if x.xValid >= VALID:
                return x.xValue
            return x.text
        else:
            if isinstance(x, ModelObject):
                e = x
            if e is not None:
                if getattr(e, "xValid", 0) == VALID_NO_CONTENT:
                    raise XPathException(p, 'err:FOTY0012', _('Atomizing element {0} that does not have a typed value').format(x))
                if e.get("{http://www.w3.org/2001/XMLSchema-instance}nil") == "true":
                    return []
                try:
                    if e.xValid >= VALID:
                        return e.xValue
                except AttributeError:
                    pass
                xModelObject = cast(ModelObject, x)
                modelXbrl = xModelObject.modelXbrl
                assert modelXbrl is not None
                modelConcept = modelXbrl.qnameConcepts.get(xModelObject.qname)
                if modelConcept is not None:
                    baseXsdType = modelConcept.baseXsdType
                else:
                    baseXsdType = "string"
                v = xModelObject.stringValue
        if baseXsdType in ("float", "double"):
            try:
                x = float(v)
            except ValueError:
                raise XPathException(p, 'err:FORG0001', _('Atomizing {0} to a {1} does not have a proper value').format(x, baseXsdType))
        elif baseXsdType == "decimal":
            try:
                x = Decimal(v)
            except InvalidOperation:
                raise XPathException(p, 'err:FORG0001', _('Atomizing {0} to decimal does not have a proper value').format(x))
        elif baseXsdType in (
            "integer",
            "nonPositiveInteger", "negativeInteger", "nonNegativeInteger", "positiveInteger",
            "long", "unsignedLong",
            "int", "unsignedInt",
            "short", "unsignedShort",
            "byte", "unsignedByte",
        ):
            try:
                x = int(v)
            except ValueError:
                raise XPathException(p, 'err:FORG0001', _('Atomizing {0} to an integer does not have a proper value').format(x))
        elif baseXsdType == "boolean":
            x = v == "true" or v == "1"
        elif baseXsdType == "QName" and e is not None:
            x = qname(e, v)
        elif baseXsdType == "anyURI":
            x = anyURI(v.strip())
        elif baseXsdType in (
            "normalizedString",
            "token",
            "language",
            "NMTOKEN",
            "Name",
            "NCName",
            "ID",
            "IDREF",
            "ENTITY",
        ):
            x = v.strip()
        elif baseXsdType == "XBRLI_DATEUNION":
            x = dateTime(v, type=DATEUNION)
        elif baseXsdType == "date":
            x = dateTime(v, type=DATE)
        elif baseXsdType == "dateTime":
            x = dateTime(v, type=DATETIME)
        elif baseXsdType in GREGORIAN_TYPES and isinstance(v, GREGORIAN_TYPES):
            x = v
        elif baseXsdType == "noContent":
            x = None  # can't be atomized
        elif baseXsdType:
            x = str(v)
        return x

    def effectiveBooleanValue(self, p: FormulaToken | None, x: ResultStack | Sequence[ContextItem] | None) -> bool:
        assert boolean is not None
        return boolean(self, p, None, [self.flattenSequence(x)])

    def traceEffectiveVariableValue(self, elt: ModelObject, varname: str) -> str | None:
        # used for tracing variable value
        if varname.startswith('$'):
            varQname = qname(elt, varname[1:])
            if varQname in self.inScopeVars:
                varValue = self.inScopeVars[varQname]
                if isinstance(varValue, ModelFact):
                    return varValue.effectiveValue
                else:
                    return str(varValue)
            else:
                return varname
        else:  # not a variable name
            return varname

    # flatten into a sequence
    def flattenSequence(self, x: Any, sequence: list[Any] | None = None) -> list[Any]:
        if sequence is None:
            if not isinstance(x, SEQUENCE_TYPES):
                if x is None:
                    return []  # none as atomic value is an empty sequence in xPath semantics
                return [x]
            sequence = []
        for el in x:
            if isinstance(el, SEQUENCE_TYPES):
                self.flattenSequence(el, sequence)
            else:
                sequence.append(el)
        return sequence

    '''  (note: slice operation makes the below slower than the above by about 15%)
    def flattenSequence(self, x):
        sequenceTypes=SEQUENCE_TYPES
        if not isinstance(x, sequenceTypes):
            return [x]
        needsFlattening = False  # no need to do anything
        for i, e in enumerate(x):
            if isinstance(e, sequenceTypes):
                needsFlattening = True # needs action at i
                break
        if needsFlattening:
            x = list(x) # start with fresh copy of list
            while i < len(x):
                if isinstance(x[i], sequenceTypes):
                    x[i:i+1] = list(x[i])
                else:
                    i += 1
        return x
    '''

    # order nodes
    def documentOrderedNodes(self, x: Iterable[ContextItem]) -> Sequence[ContextItem]:
        l = set()  # must have unique nodes only
        for e in x:
            if isinstance(e, ModelObject):
                h = cast(int, e.sourceline)
            elif isinstance(e, ModelAttribute):
                h = cast(int, e.modelElement.sourceline)
            else:
                h = 0
            l.add((h, e))
        return [e for h, e in sorted(l, key=lambda _h: _h[0] or 0)]  # or 0 in case sourceline is None

    def modelItem(self, x: ModelObject) -> ModelFact | None:
        if isinstance(x, (ModelFact, ModelInlineFact)) and x.isItem:
            return x
        return None

    def modelInstance(self, x: ModelXbrl | ModelObject | None) -> ModelXbrl | None:
        if isinstance(x, ModelXbrl):
            return x
        if isinstance(x, ModelObject):
            return x.modelXbrl
        return None



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations
from typing import Any

from arelle.ModelDocument import LoadingException, ModelDocument
from arelle.ModelValue import qname
from arelle.ModelXbrl import ModelXbrl
from arelle.ValidateXbrl import ValidateXbrl
from arelle.typing import TypeGetText
from arelle.utils.validate.ValidationPlugin import ValidationPlugin
from .DisclosureSystems import DISCLOSURE_SYSTEM_NT16, DISCLOSURE_SYSTEM_NT17, DISCLOSURE_SYSTEM_NT18
from .PluginValidationDataExtension import PluginValidationDataExtension

_: TypeGetText


class ValidationPluginExtension(ValidationPlugin):
    def newPluginData(self, validateXbrl: ValidateXbrl) -> PluginValidationDataExtension:
        disclosureSystem = validateXbrl.disclosureSystem.name
        if disclosureSystem == DISCLOSURE_SYSTEM_NT16:
            jenvNamespace = 'http://www.nltaxonomie.nl/nt16/jenv/20211208/dictionary/jenv-bw2-data'
            kvkINamespace = 'http://www.nltaxonomie.nl/nt16/kvk/20211208/dictionary/kvk-data'
            nlTypesNamespace = 'http://www.nltaxonomie.nl/nt16/sbr/20210301/dictionary/nl-types'
            entrypointRoot = 'http://www.nltaxonomie.nl/nt16/kvk/20211208/entrypoints/'
            entrypoints = {entrypointRoot + e for e in [
                'kvk-rpt-jaarverantwoording-2021-ifrs-full.xsd',
                'kvk-rpt-jaarverantwoording-2021-ifrs-geconsolideerd-nlgaap-enkelvoudig.xsd',
                'kvk-rpt-jaarverantwoording-2021-ifrs-smes.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-banken.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-beleggingsentiteiten.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-cooperaties.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-cv-vof.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-fondsenwervende-organisaties-klein.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-fondsenwervende-organisaties.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-groot-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-groot.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-klein-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-klein-verticaal-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-klein-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-klein.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-micro-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-micro.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-middelgroot-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-middelgroot-verticaal-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-middelgroot-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-middelgroot.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-organisaties-zonder-winststreven-klein.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-organisaties-zonder-winststreven.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-pensioenfondsen.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-premiepensioeninstellingen.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-stichtingen.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-toegelaten-instellingen-volkshuisvesting.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-verzekeringsmaatschappijen.xsd',
                'kvk-rpt-jaarverantwoording-2021-nlgaap-zorginstellingen.xsd',
            ]}
        elif disclosureSystem == DISCLOSURE_SYSTEM_NT17:
            jenvNamespace = 'http://www.nltaxonomie.nl/nt17/jenv/20221214/dictionary/jenv-bw2-data'
            kvkINamespace = 'http://www.nltaxonomie.nl/nt17/kvk/20221214/dictionary/kvk-data'
            nlTypesNamespace = 'http://www.nltaxonomie.nl/nt17/sbr/20220301/dictionary/nl-types'
            entrypointRoot = 'http://www.nltaxonomie.nl/nt17/kvk/20221214/entrypoints/'
            entrypoints = {entrypointRoot + e for e in [
                'kvk-rpt-jaarverantwoording-2022-ifrs-full.xsd',
                'kvk-rpt-jaarverantwoording-2022-ifrs-geconsolideerd-nlgaap-enkelvoudig.xsd',
                'kvk-rpt-jaarverantwoording-2022-ifrs-smes.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-banken.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-beleggingsentiteiten.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-cooperaties.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-cv-vof.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-fondsenwervende-organisaties-klein.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-fondsenwervende-organisaties.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-groot-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-groot.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-klein-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-klein-verticaal-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-klein-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-klein.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-micro-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-micro.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-middelgroot-plus.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-middelgroot-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-middelgroot-verticaal-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-middelgroot-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-middelgroot.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-organisaties-zonder-winststreven-klein.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-organisaties-zonder-winststreven.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-pensioenfondsen.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-premiepensioeninstellingen.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-stichtingen.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-toegelaten-instellingen-volkshuisvesting.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-verzekeringsmaatschappijen.xsd',
                'kvk-rpt-jaarverantwoording-2022-nlgaap-zorginstellingen.xsd',
            ]}
        elif disclosureSystem == DISCLOSURE_SYSTEM_NT18:
            jenvNamespace = 'http://www.nltaxonomie.nl/nt18/jenv/20231213/dictionary/jenv-bw2-data'
            kvkINamespace = 'http://www.nltaxonomie.nl/nt18/kvk/20231213/dictionary/kvk-data'
            nlTypesNamespace = 'http://www.nltaxonomie.nl/nt18/sbr/20230301/dictionary/nl-types'
            entrypointRoot = 'http://www.nltaxonomie.nl/nt18/kvk/20231213/entrypoints/'
            entrypoints = {entrypointRoot + e for e in [
                'kvk-rpt-jaarverantwoording-2023-ifrs-full.xsd',
                'kvk-rpt-jaarverantwoording-2023-ifrs-geconsolideerd-nlgaap-enkelvoudig.xsd',
                'kvk-rpt-jaarverantwoording-2023-ifrs-smes.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-banken.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-beleggingsentiteiten.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-cooperaties.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-cv-vof.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-fondsenwervende-organisaties-klein.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-fondsenwervende-organisaties.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-groot-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-groot.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-klein-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-klein-verticaal-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-klein-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-klein.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-micro-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-micro.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-middelgroot-plus.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-middelgroot-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-middelgroot-verticaal-publicatiestukken.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-middelgroot-verticaal.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-middelgroot.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-organisaties-zonder-winststreven-klein.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-organisaties-zonder-winststreven.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-pensioenfondsen.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-premiepensioeninstellingen.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-stichtingen.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-toegelaten-instellingen-volkshuisvesting.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-verzekeringsmaatschappijen.xsd',
                'kvk-rpt-jaarverantwoording-2023-nlgaap-zorginstellingen.xsd',
            ]}
        else:
            raise ValueError(f'Invalid NL disclosure system: {disclosureSystem}')
        return PluginValidationDataExtension(
            self.name,
            financialReportingPeriodCurrentStartDateQn=qname(f'{{{jenvNamespace}}}FinancialReportingPeriodCurrentStartDate'),
            financialReportingPeriodCurrentEndDateQn=qname(f'{{{jenvNamespace}}}FinancialReportingPeriodCurrentEndDate'),
            financialReportingPeriodPreviousStartDateQn=qname(f'{{{jenvNamespace}}}FinancialReportingPeriodPreviousStartDate'),
            financialReportingPeriodPreviousEndDateQn=qname(f'{{{jenvNamespace}}}FinancialReportingPeriodPreviousEndDate'),
            formattedExplanationItemTypeQn=qname(f'{{{nlTypesNamespace}}}formattedExplanationItemType'),
            documentAdoptionDateQn=qname(f'{{{jenvNamespace}}}DocumentAdoptionDate'),
            documentAdoptionStatusQn=qname(f'{{{jenvNamespace}}}DocumentAdoptionStatus'),
            documentResubmissionUnsurmountableInaccuraciesQn=qname(f'{{{kvkINamespace}}}DocumentResubmissionDueToUnsurmountableInaccuracies'),
            entrypointRoot=entrypointRoot,
            entrypoints=entrypoints,
            textFormattingSchemaPath='sbr-text-formatting.xsd',
            textFormattingWrapper='<formattedText xmlns="http://www.nltaxonomie.nl/2017/xbrl/sbr-text-formatting">{}</formattedText>',
        )

    def modelXbrlLoadComplete(self, modelXbrl: ModelXbrl, *args: Any, **kwargs: Any) -> ModelDocument | LoadingException | None:
        if self.disclosureSystemFromPluginSelected(modelXbrl):
            disclosureSystem = modelXbrl.modelManager.disclosureSystem.name
            if disclosureSystem in (DISCLOSURE_SYSTEM_NT16, DISCLOSURE_SYSTEM_NT17, DISCLOSURE_SYSTEM_NT18):
                # Dutch taxonomies prior to 2025 incorrectly used hypercube linkrole for roots instead of dimension linkrole.
                paramQName = qname('tlbDimRelsUseHcRoleForDomainRoots', noPrefixIsNoNamespace=True)
                modelXbrl.modelManager.formulaOptions.parameterValues[paramQName] = (None, "true")
        return None



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import inspect
from pathlib import Path
from types import ModuleType
from typing import Any

from arelle.DisclosureSystem import DisclosureSystem
from arelle.ModelDocument import LoadingException, ModelDocument
from arelle.ModelXbrl import ModelXbrl
from arelle.ValidateXbrl import ValidateXbrl
from arelle.utils.PluginHooks import ValidationHook
from arelle.utils.validate.Decorator import (
    ValidationAttributes,
    ValidationFunction,
    getValidationAttributes,
)
from arelle.utils.PluginData import PluginData


class ValidationPlugin:
    def __init__(
        self,
        name: str,
        disclosureSystemConfigUrl: Path,
        validationTypes: list[str],
        validationRuleModules: list[ModuleType],
    ) -> None:
        """
        Base validation plugin class. Can be initialized directly, or extended if you require additional plugin hooks.
        This class is intended to be used in conjunction with the [@validation](#arelle.utils.validate.Decorator.validation) decorator.

        If you need to cache or store data between rules, you should implement a dataclass that extends [PluginData](#arelle.utils.PluginData.PluginData)
        and override the [newPluginData](#newPluginData) method to return your dataclass.

        :param disclosureSystemConfigUrl: A path to the plugin disclosure system xml config file.
        :param validationTypes: A list of validation types for the plugin. These should correspond to the validation types
               in your disclosure system xml file.
        :param validationRulesModule: The rules module which will be searched recursively for rule functions decorated with @validation.
        """
        self.name = name
        self._disclosureSystemConfigURL = disclosureSystemConfigUrl
        self._validationTypes = tuple(validationTypes)
        if not validationRuleModules:
            raise ValueError("At least one rules module must be provided.")
        self._validationModules = validationRuleModules
        self._validationsDiscovered = False
        self._rulesByDisclosureSystemByHook: dict[
            ValidationHook, dict[str, list[ValidationFunction]]
        ] = {}
        self._excludedDisclosureSystemsByRulesByHook: dict[
            ValidationHook, dict[ValidationFunction, set[str]]
        ] = {}

    def newPluginData(self, validateXbrl: ValidateXbrl) -> PluginData:
        """
        Returns a dataclass intended to be overriden by plugins to facilitate caching and passing data between rule functions.
        The default implementation doesn't provide any fields other than the plugin name.

        :return: An instance of PluginData.
        """
        return PluginData(self.name)

    @property
    def validationTypes(self) -> tuple[str, ...]:
        return self._validationTypes

    @property
    def disclosureSystemConfigURL(self) -> str:
        return str(self._disclosureSystemConfigURL)

    @property
    def disclosureSystemTypes(self) -> tuple[tuple[str, str], ...]:
        return tuple(
            (validationType, f"pluginValidationType{validationType}")
            for validationType in self.validationTypes
        )

    def modelDocumentPullLoader(
        self,
        modelXbrl: ModelXbrl,
        normalizedUri: str,
        filepath: str,
        isEntry: bool,
        namespace: str | None,
        *args: Any,
        **kwargs: Any,
    ) -> ModelDocument | LoadingException | None:
        raise NotImplementedError

    def modelXbrlLoadComplete(
        self,
        modelXbrl: ModelXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> ModelDocument | LoadingException | None:
        raise NotImplementedError

    def validateXbrlStart(
        self,
        val: ValidateXbrl,
        parameters: dict[Any, Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Executes validation functions in the rules module that was provided to the constructor of this class.
        Each function decorated with [@validation](#arelle.utils.validate.Decorator.validation) will be run if:
        1. the decorator was used with the xbrl start hook: `@validation(hook=ValidationHook.XBRL_START)`
        2. the user selected disclosure system is in line with the combination of disclosure systems provided to the decorator.

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param parameters: dictionary of validation parameters.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        self._executeValidations(ValidationHook.XBRL_START, val, parameters, *args, **kwargs)

    def validateXbrlFinally(
        self,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Executes validation functions in the rules module that was provided to the constructor of this class.
        Each function decorated with [@validation](#arelle.utils.validate.Decorator.validation) will be run if:
        1. the decorator was used with the xbrl finally hook: `@validation(hook=ValidationHook.XBRL_FINALLY)`
        2. the user selected disclosure system is in line with the combination of disclosure systems provided to the decorator.

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        self._executeValidations(ValidationHook.XBRL_FINALLY, val, *args, **kwargs)

    def validateXbrlDtsDocument(
        self,
        val: ValidateXbrl,
        modelDocument: ModelDocument,
        isFilingDocument: bool,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Executes validation functions in the rules module that was provided to the constructor of this class.
        Each function decorated with [@validation](#arelle.utils.validate.Decorator.validation) will be run if:
        1. the decorator was used with the xbrl dts document hook: `@validation(hook=ValidationHook.XBRL_DTS_DOCUMENT)`
        2. the user selected disclosure system is in line with the combination of disclosure systems provided to the decorator.

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        self._executeValidations(ValidationHook.XBRL_DTS_DOCUMENT, val, modelDocument, isFilingDocument, *args, **kwargs)

    def validateFinally(
        self,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Executes validation functions in the rules module that was provided to the constructor of this class.
        Each function decorated with [@validation](#arelle.utils.validate.Decorator.validation) will be run if:
        1. the decorator was used with the validate finally hook: `@validation(hook=ValidationHook.FINALLY)`
        2. the user selected disclosure system is in line with the combination of disclosure systems provided to the decorator.

        :param val: The [ValidateXBRL](#arelle.ValidateXbrl.ValidateXbrl) instance.
        :param args: Argument capture to ensure new parameters don't break plugin hook.
        :param kwargs: Argument capture to ensure new named parameters don't break plugin hook.
        :return: None
        """
        self._executeValidations(ValidationHook.FINALLY, val, *args, **kwargs)

    def _executeValidations(
        self,
        pluginHook: ValidationHook,
        validateXbrl: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if self.disclosureSystemFromPluginSelected(validateXbrl):
            pluginData = validateXbrl.getPluginData(self.name)
            if pluginData is None:
                pluginData = self.newPluginData(validateXbrl)
                validateXbrl.setPluginData(pluginData)
            for rule in self._getValidations(validateXbrl.disclosureSystem, pluginHook):
                validations = rule(pluginData, validateXbrl, *args, **kwargs)
                if validations is not None:
                    modelXbrl = validateXbrl.modelXbrl
                    for val in validations:
                        modelXbrl.log(level=val.level.name, codes=val.codes, msg=val.msg, **val.args)

    def disclosureSystemFromPluginSelected(
        self,
        model: ValidateXbrl | ModelXbrl,
    ) -> bool:
        if isinstance(model, ValidateXbrl):
            disclosureSystem = model.disclosureSystem
        elif isinstance(model, ModelXbrl):
            disclosureSystem = model.modelManager.disclosureSystem
        else:
            raise ValueError(f"Expected param to be either ValidateXbrl or ModelXbrl, received {type(model)}.")
        return any(
            getattr(disclosureSystem, testProperty, False)
            for _, testProperty in self.disclosureSystemTypes
        )

    def _getValidations(
        self,
        disclosureSystem: DisclosureSystem,
        pluginHook: ValidationHook,
    ) -> list[ValidationFunction]:
        if disclosureSystem.name is None:
            raise RuntimeError("Disclosure system not configured.")
        if not self._validationsDiscovered:
            self._discoverValidations()
        nonExcludedRules = [
            func
            for func, excluded in self._excludedDisclosureSystemsByRulesByHook.get(pluginHook, {}).items()
            if disclosureSystem.name not in excluded
        ]
        includedRules = self._rulesByDisclosureSystemByHook.get(pluginHook, {}).get(disclosureSystem.name, [])
        return nonExcludedRules + includedRules

    def _discoverValidations(self) -> None:
        rulesFunctionDiscovered = False
        for mod in self._validationModules:
            for _, func in inspect.getmembers(mod, inspect.isfunction):
                attributes = getValidationAttributes(func)
                if attributes is not None:
                    rulesFunctionDiscovered = True
                    self._storeValidationFunction(func, attributes)
        if not rulesFunctionDiscovered:
            raise RuntimeError(f"No @validation rules found for '{self.name}' validation plugin.")
        self._validationsDiscovered = True

    def _storeValidationFunction(
        self,
        func: ValidationFunction,
        attributes: ValidationAttributes,
    ) -> None:
        hook = attributes.hook
        disclosureSystems = attributes.disclosureSystems
        if disclosureSystems is None:
            excludedDisclosureSystemsByRules = self._excludedDisclosureSystemsByRulesByHook.setdefault(hook, {})
            excludedDisclosureSystems = excludedDisclosureSystemsByRules.setdefault(func, set())
            excludedDisclosureSystems.update(attributes.excludeDisclosureSystems)
        else:
            rulesByDisclosureSystem = self._rulesByDisclosureSystemByHook.setdefault(hook, {})
            for disclosureSystem in disclosureSystems:
                rulesByDisclosureSystem.setdefault(disclosureSystem, []).append(func)



# -*- coding: utf-8 -*-

'''
UKCompaniesHouseLoader is a plug-in to both GUI menu and command line/web service
that loads a Companies House zip XBRL instances file.

See COPYRIGHT.md for copyright information.
'''
from lxml import html
import datetime, os
from arelle import FileSource
from arelle.ModelRssObject import ModelRssObject
from arelle.Version import authorLabel, copyrightLabel
from arelle.XmlValidate import UNVALIDATED, VALID

class CompaniesHouseItem:
    def __init__(self, modelXbrl, fileName, entryUrl):
        self.cikNumber = None
        self.accessionNumber = fileName[4:].partition(".")[0]
        self.fileNumber = None
        self.companyName = None
        self.formType = None
        pubDate = fileName.rpartition("_")[2].partition(".")[0]
        try:
            self.pubDate = datetime.datetime(int(pubDate[0:4]), int(pubDate[4:6]), int(pubDate[6:8]))
            self.acceptanceDatetime = self.pubDate
            self.filingDate = self.pubDate.date()
        except ValueError:
            self.pubDate = self.acceptanceDatetime = self.filingDate = None
        self.period = None
        self.assignedSic = None
        self.fiscalYearEnd = None
        self.htmlUrl = None
        self.url = entryUrl
        self.zippedUrl = entryUrl
        self.htmURLs = ((fileName, entryUrl),)
        self.status = "not tested"
        self.results = None
        self.assertions = None
        self.objectIndex = len(modelXbrl.modelObjects)
        modelXbrl.modelObjects.append(self)

    def setResults(self, modelXbrl):
        self.results = []
        self.assertionUnsuccessful = False
        # put error codes first, sorted, then assertion result (dict's)
        self.status = "pass"
        for error in modelXbrl.errors:
            if isinstance(error,dict):  # assertion results
                self.assertions = error
                for countSuccessful, countNotsuccessful in error.items():
                    if countNotsuccessful > 0:
                        self.assertionUnsuccessful = True
                        self.status = "unsuccessful"
            else:   # error code results
                self.results.append(error)
                self.status = "fail" # error code
        self.results.sort()

    def objectId(self,refId=""):
        """Returns a string surrogate representing the object index of the model document,
        prepended by the refId string.
        :param refId: A string to prefix the refId for uniqueless (such as to use in tags for tkinter)
        :type refId: str
        """
        return "_{0}_{1}".format(refId, self.objectIndex)

def companiesHouseInstanceLoaded(modelXbrl, rssWatchOptions, rssItem, *args, **kwargs):
    for fact in modelXbrl.factsInInstance:
        name = fact.qname.localName if fact.qname is not None else None
        if name in ("CompaniesHouseRegisteredNumber",
                    "UKCompaniesHouseRegisteredNumber",
                    "EntityCurrentLegalName",
                    "EntityCurrentLegalOrRegisteredName",
                    "EndDateForPeriodCoveredByReport"):
            if name in ("CompaniesHouseRegisteredNumber","UKCompaniesHouseRegisteredNumber"):
                rssItem.cikNumber = fact.value.strip()
            elif name in ("EntityCurrentLegalName", "EntityCurrentLegalOrRegisteredName"):
                rssItem.companyName = fact.value.strip()
            elif name == "EndDateForPeriodCoveredByReport" and getattr(fact, "xValid", UNVALIDATED) == VALID:
                rssItem.period = fact.xValue
                rssItem.fiscalYearEnd = "{:04}-{:02}".format(fact.xValue.year, fact.xValue.month)


def companiesHouseLoader(modelXbrl, mappedUri, filepath, *args, **kwargs):
    if not (mappedUri.startswith("http://download.companieshouse.gov.uk/") and
            mappedUri.endswith(".zip")):
        return None # not a companies houst zip file

    rssObject = ModelRssObject(modelXbrl, uri=mappedUri, filepath=filepath)

    # find <table> with <a>Download in it
    for instanceFile in modelXbrl.fileSource.dir:
        rssObject.rssItems.append(
            CompaniesHouseItem(modelXbrl, instanceFile, mappedUri + '/' + instanceFile))
    return rssObject



__pluginInfo__ = {
    'name': 'UK Companies House Loader',
    'version': '0.9',
    'description': "This plug-in loads UK Companies House XBRL documents.  ",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    # take out for now: 'CntlrCmdLine.Options': streamingOptionsExtender,
    'ModelDocument.PullLoader': companiesHouseLoader,
    'RssItem.Xbrl.Loaded': companiesHouseInstanceLoaded,
}



'''
Hello dolly is a simple "Hello world" to demonstrate how plug-ins
are written for Arelle

See COPYRIGHT.md for copyright information.
'''
from arelle.Version import copyrightLabel

def menuEntender(cntlr, menu):
    menu.add_command(label="Hello i18n", underline=0, command=lambda: menuCommand(cntlr) )

def menuCommand(cntlr):
    i10L_world = _("Hello World");
    cntlr.addToLog(i10L_world)
    import tkinter
    tkinter.messagebox.showinfo(_("Prints 'Hello World'"), i10L_world, parent=cntlr.parent)

'''
   Do not use _( ) in pluginInfo itself (it is applied later, after loading
'''
__pluginInfo__ = {
    'name': 'Hello i18n',
    'version': '0.9',
    'description': '''Minimal plug-in that demonstrates i18n internationalization by localized gettext.''',
    'localeURL': "locale",
    'localeDomain': 'hello_i18n',
    'license': 'Apache-2',
    'author': 'R\u00e9gis D\u00e9camps',
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrWinMain.Menu.Tools': menuEntender
}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from typing import Iterable, cast

from arelle.ModelInstanceObject import ModelFact
from arelle.ModelXbrl import ModelXbrl
from arelle.XmlValidateConst import VALID
from arelle.utils.validate.Validation import Validation


def errorOnMissingRequiredFact(
        modelXbrl: ModelXbrl,
        conceptLn: str,
        code: str,
        message: str,
) -> Iterable[Validation]:
    """
    Yields an error if a fact with the given localName is not tagged with a non-nil value.
    :return: Yields validation errors.
    """
    facts: set[ModelFact] = modelXbrl.factsByLocalName.get(conceptLn, set())
    for fact in facts:
        if not fact.isNil:
            return
    yield Validation.error(
        conceptLn=conceptLn,
        codes=code,
        msg=message,
    )


def errorOnNegativeFact(
        modelXbrl: ModelXbrl,
        conceptLn: str,
        code: str,
        message: str,
) -> Iterable[Validation]:
    """
    Yields an error if a fact with the given localName is tagged with a negative value.
    :return: Yields validation errors.
    """
    errorModelFacts: list[ModelFact] = []
    facts = modelXbrl.factsByLocalName.get(conceptLn, set())
    for fact in facts:
        if fact.xValid >= VALID and cast(int, fact.xValue) < 0:
            errorModelFacts.append(fact)
    if errorModelFacts:
        yield Validation.error(
            codes=code,
            modelObject=errorModelFacts,
            msg=message,
        )



'''
See COPYRIGHT.md for copyright information.
'''
import os

from arelle.Version import authorLabel, copyrightLabel
from .....plugin.validate.EFM.Document import checkDTSdocument
from .....plugin.validate.EFM.Filing import validateFiling


def dislosureSystemTypes(disclosureSystem, *args, **kwargs):
    # return ((disclosure system name, variable name), ...)
    return (("GFM", "GFMplugin"),)

def disclosureSystemConfigURL(disclosureSystem, *args, **kwargs):
    return os.path.join(os.path.dirname(__file__), "config.xml")

def validateXbrlStart(val, parameters=None, *args, **kwargs):
    val.validateGFMplugin = val.validateDisclosureSystem and getattr(val.disclosureSystem, "GFMplugin", False)
    if not (val.validateGFMplugin):
        return

def validateXbrlFinally(val, *args, **kwargs):
    if not (val.validateGFMplugin):
        return

    modelXbrl = val.modelXbrl

    _statusMsg = _("validating {0} filing rules").format(val.disclosureSystem.name)
    modelXbrl.profileActivity()
    modelXbrl.modelManager.showStatus(_statusMsg)

    validateFiling(val, modelXbrl, isGFM=True)

    modelXbrl.profileActivity(_statusMsg, minTimeToShow=0.0)
    modelXbrl.modelManager.showStatus(None)

def validateXbrlDtsDocument(val, modelDocument, isFilingDocument, *args, **kwargs):
    if not (val.validateGFMplugin) or not modelDocument:
        return

    checkDTSdocument(val, modelDocument, isFilingDocument)


__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Validate GFM',
    'version': '0.9',
    'description': '''GFM Validation.''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'DisclosureSystem.Types': dislosureSystemTypes,
    'DisclosureSystem.ConfigURL': disclosureSystemConfigURL,
    'Validate.XBRL.Start': validateXbrlStart,
    'Validate.XBRL.Finally': validateXbrlFinally,
    'Validate.XBRL.DTS.document': validateXbrlDtsDocument,
}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from arelle.utils.PluginData import PluginData


class PluginValidationDataExtension(PluginData):
    positiveFactConcepts: set[str] | None = None



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import json
import os
from collections import defaultdict
from collections.abc import Collection
from typing import Any, Dict, Generator, List, Union, cast
import regex as re

from lxml.etree import _Element

from arelle.FileSource import openFileStream
from arelle.ModelDocument import ModelDocument
from arelle.ModelDtsObject import ModelConcept
from arelle.ModelInstanceObject import ModelContext, ModelFact, ModelUnit
from arelle.ModelManager import ModelManager
from arelle.ModelObject import ModelObject
from arelle.ModelRelationshipSet import ModelRelationshipSet
from arelle.ModelValue import QName
from arelle.ModelXbrl import ModelXbrl
from arelle.ValidateXbrl import ValidateXbrl
from arelle.XmlValidateConst import VALID
from arelle.typing import TypeGetText
from .Const import esefCorNsPattern, esefNotesStatementConcepts, esefTaxonomyNamespaceURIs, htmlEventHandlerAttributes, svgEventAttributes

_: TypeGetText

YEAR_GROUP = "year"
DISCLOSURE_SYSTEM_YEAR_PATTERN = re.compile(rf"esef-(?:unconsolidated-)?(?P<{YEAR_GROUP}>20\d\d)")


def getDisclosureSystemYear(modelXbrl: ModelXbrl) -> int:
    for name in modelXbrl.modelManager.disclosureSystem.names:
        disclosureSystemYear = DISCLOSURE_SYSTEM_YEAR_PATTERN.fullmatch(name)
        if disclosureSystemYear:
            return int(disclosureSystemYear.group(YEAR_GROUP))
    raise ValueError(f"Unable to determine year of ESEF disclosure system matching pattern 'esef-20XX' from {modelXbrl.modelManager.disclosureSystem.names}")


# check if a modelDocument URI is an extension URI (document URI)
# also works on a uri passed in as well as modelObject
def isExtension(val: ValidateXbrl, modelObject: ModelObject | ModelDocument | str | None) -> bool:
    if modelObject is None:
        return False
    if isinstance(modelObject, str):
        uri = modelObject
    else:
        uri = modelObject.modelDocument.uri
    return (uri.startswith(val.modelXbrl.uriDir) or
            not any(uri.startswith(standardTaxonomyURI) for standardTaxonomyURI in val.authParam["standardTaxonomyURIs"]))


# check if in core esef taxonomy (based on namespace URI)
def isInEsefTaxonomy(val: ValidateXbrl, modelObject: ModelObject | None) -> bool:
    if modelObject is None:
        return False
    assert modelObject.qname is not None
    ns = modelObject.qname.namespaceURI
    assert ns is not None
    return (any(ns.startswith(esefNsPrefix) for esefNsPrefix in esefTaxonomyNamespaceURIs))


def resourcesFilePath(modelManager: ModelManager, fileName: str) -> str:
    # resourcesDir can be in cache dir (production) or in validate/EFM/resources (for development)
    _resourcesDir = os.path.join( os.path.dirname(__file__), "resources") # dev/testing location

    if not os.path.isabs(_resourcesDir):
        _resourcesDir = os.path.abspath(_resourcesDir)
    if not os.path.exists(_resourcesDir): # production location
        _resourcesDir = os.path.join(modelManager.cntlr.webCache.cacheDir, "resources", "validation", "ESEF")

    return os.path.join(_resourcesDir, fileName)


def loadAuthorityValidations(modelXbrl: ModelXbrl) -> list[Any] | dict[Any, Any]:
    _file = openFileStream(modelXbrl.modelManager.cntlr, resourcesFilePath(modelXbrl.modelManager, "authority-validations.json"), 'rt', encoding='utf-8')
    validations = json.load(_file) # {localName: date, ...}
    _file.close()
    return cast(Union[Dict[Any, Any], List[Any]], validations)


def checkForMultiLangDuplicates(modelXbrl: ModelXbrl) -> None:
    _factConceptContextUnitHash: defaultdict[int, list[ModelFact]] = defaultdict(list)

    for f in modelXbrl.factsInInstance:
        if (
            (f.isNil or getattr(f, "xValid", 0) >= VALID)
            and f.context is not None
            and f.concept is not None
            and f.concept.type is not None
            and f.concept.type.isWgnStringFactType
        ):
            _factConceptContextUnitHash[f.conceptContextUnitHash].append(f)

    for hashEquivalentFacts in _factConceptContextUnitHash.values():
        if len(hashEquivalentFacts) <= 1:  # skip facts present only once
            continue
        _aspectEqualFacts: defaultdict[tuple[QName, str], dict[tuple[ModelContext, ModelUnit | None], list[ModelFact]]] = defaultdict(dict)
        for f in hashEquivalentFacts:  # check for hash collision by value checks on context and unit
            cuDict = _aspectEqualFacts[(f.qname, (f.xmlLang or "").lower())]
            _matched = False
            for (_cntx, _unit), fList in cuDict.items():
                if (f.context.isEqualTo(_cntx)
                        and ((_unit is None and f.unit is None)
                             or (f.unit is not None and f.unit.isEqualTo(_unit)))):
                    _matched = True
                    fList.append(f)
                    break
            if not _matched:
                cuDict[(f.context, f.unit)] = [f]
        for cuDict in _aspectEqualFacts.values():
            for fList in cuDict.values():
                if len(fList) > 1 and not all(f.xValue == fList[0].xValue for f in fList):
                    modelXbrl.warning("ESEF.2.2.4.inconsistentDuplicateNonnumericFactInInlineXbrlDocument",
                        "Inconsistent duplicate non-numeric facts SHOULD NOT appear in the content of an inline XBRL document. "
                        "%(fact)s that was used more than once in contexts equivalent to %(contextID)s, with different values but same language (%(language)s).",
                        modelObject=fList, fact=fList[0].qname, contextID=fList[0].contextID, language=fList[0].xmlLang)


def getEsefNotesStatementConcepts(modelXbrl: ModelXbrl) -> set[str]:
    document_name_spaces = modelXbrl.namespaceDocs
    esef_notes_statement_concepts:set[str] = set()
    esef_cor_Nses = []
    for targetNs, models in document_name_spaces.items():
        if esefCorNsPattern.match(targetNs):
            found_prefix = ''
            found_namespace = ''
            for prefix, namespace in models[0].targetXbrlRootElement.nsmap.items():
                if targetNs == namespace:
                    found_namespace = targetNs
                    found_prefix = '' if prefix is None else prefix
                    break
            esef_cor_Nses.append((found_prefix, found_namespace))
    if len(esef_cor_Nses) == 0:
        modelXbrl.error("ESEF.RTS.efrsCoreRequired",
                          _("RTS on ESEF requires EFRS core taxonomy."),
                          modelObject=modelXbrl)
    elif len(esef_cor_Nses) > 1:
        modelXbrl.warning("Arelle.ESEF.multipleEsefTaxonomies",
                        _("Multiple ESEF taxonomies were imported %(esefNamespaces)s."),
                        modelObject=modelXbrl, esefNamespaces=", ".join(ns[1] for ns in esef_cor_Nses))
    else:
        esef_notes_statement_concepts = set(str(QName(esef_cor_Nses[0][0], esef_cor_Nses[0][1], n)) for n in esefNotesStatementConcepts)
    return esef_notes_statement_concepts


def isChildOfNotes(
    child: ModelConcept,
    relSet: ModelRelationshipSet,
    esefNotesConcepts: set[str],
    _visited: set[ModelConcept],
) -> bool:
    if len(esefNotesConcepts) == 0:
        return False
    relations_to = relSet.toModelObject(child)
    if not relations_to and str(child.qname) in esefNotesConcepts:
        return True

    _visited.add(child)
    for rel in relations_to:
        parent = rel.fromModelObject
        if parent is not None and parent not in _visited:
            if isChildOfNotes(parent, relSet, esefNotesConcepts, _visited):
                return True
    _visited.remove(child)
    return False


def hasEventHandlerAttributes(elt: Any) -> bool:
    return _hasEventAttributes(elt, htmlEventHandlerAttributes)


def hasSvgEventAttributes(elt: Any) -> bool:
    return _hasEventAttributes(elt, svgEventAttributes)


def _hasEventAttributes(elt: Any, attributes: Collection[str]) -> bool:
    if isinstance(elt, _Element):
        return any(a in attributes for a in elt.keys())
    return False


def etreeIterWithDepth(
    node: ModelObject | _Element,
    depth: int = 0,
) -> Generator[tuple[ModelObject | _Element, int], None, None]:
    yield node, depth
    for child in node.iterchildren():
        for n_d in etreeIterWithDepth(child, depth + 1):
            yield n_d
